from __future__ import annotations

import os

import matplotlib as mpl
import numpy as np
from ftag import Label
from ftag.hdf5 import H5Reader

from puma.utils import get_good_colours, get_good_linestyles
from puma.var_vs_var import VarVsVar, VarVsVarPlot


def n_tracks_per_origin(
    flavour_list: list[Label],
    files: dict[dict[str]],
    plot_type: str,
    plot_path: str | None = None,
    track_origin_dict: dict | None = None,
    plot_format: str = "pdf",
    plot_name: str | None = None,
    **kwargs,
):
    """Plot number of tracks per track origin plot as a function of pT.

    Parameters
    ----------
    flavour_list : list[Flavour]
        List of Flavour objects to consider.
    files : dict[dict[str]]
        Dict of dicts with the samples to use. Each sub-dict has the following variables:
        filepath : str
            Path to the file.
        process_label : str
            Label of the process stored in the file for the legend.
        pt_bins : np.ndarray
            Numpy array with the pT bins to use.
        n_jets : int, optional
            Number of jets to load. By default load all jets.
        jets_name : str, optional
            Name of the jet collection in the h5 files. By default "jets".
        tracks_name : str, optional
            Name of the track collection in the h5 files. By default "tracks".
        jet_pt_variable : str, optional
            Name of the jet pT variable in the jet collection. By default "pt".
        track_truth_variable : str, optional
            Name of the track truth origin variable. By default "ftagTruthOriginLabel".
        flavour_label_variable : str, optional
            Name of the flavour label variable. By default "HadronConeExclTruthLabelID".
    plot_type : str
        Decide, which type of plot is wanted. Supported are "all_samples_one_flavour"
        or "one_sample_all_flavour".
    plot_path : str, optional
        Path to the folder where the plots will be stored. If None, plot is saved to current
        directory. By default None
    track_origin_dict : dict | None, optional
        Dict with the track origin names as keys and their origin value(s) as a list.
        With this, you can combine different origins and give them a new name. See the
        definition of the default track_origin_dict in the code. By default None
    plot_format : str, optional
        Plot format, by default "pdf"
    plot_name : str, optional
        Extra for the plot name, by default None

    Raises
    ------
    ValueError
        If the given plot_type is not supported.
    """
    # Check which type of plot is wanted
    if plot_type.lower() == "all_samples_one_flavour":
        all_flav_plot = False

    elif plot_type.lower() == "one_sample_all_flavour":
        all_flav_plot = True

    else:
        raise ValueError(
            'Only "all_samples_one_flavour" and "one_sample_all_flavour" are supported!'
        )

    # Get the grouping of the track origins
    if track_origin_dict is None:
        track_origin_dict = {
            "All": range(8),
            "HF decay": [3, 4, 5],
            # "Fragmentation": [1, 2],
            "From $\\tau$": [6],
            "Others": [0, 1, 2, 7],
        }

    # Init a default kwargs dict for the VarVsVarPlot
    VarVsVarPlot_kwargs = {
        "xlabel": "$p_T$ [GeV]",
        "ylabel": "Average $N_\\mathrm{Tracks}$ per Origin",
        "figsize": (7.0, 4.5),
        "y_scale": 1.3,
        "atlas_first_tag": "Simulation Internal",
        "atlas_second_tag": None,
        "logx": False,
        "logy": False,
    }

    # If kwargs are given, update the VarVsVarPlot_kwargs dict
    if kwargs is not None:
        VarVsVarPlot_kwargs.update(kwargs)

    # Init a dict for the plots
    var_plot_dict = {}

    # Check plot type and init plot if needed
    if all_flav_plot:
        for file_key, file_value in files.items():
            unmodified_second_tag = VarVsVarPlot_kwargs.get("atlas_second_tag")

            # Modify the second tag
            if unmodified_second_tag is None:
                VarVsVarPlot_kwargs["atlas_second_tag"] = file_value["process_label"] + " Jets"

            else:
                VarVsVarPlot_kwargs["atlas_second_tag"] += f'\n{file_value["process_label"]} Jets'

            # Init the plot
            var_plot_dict[file_key] = VarVsVarPlot(**VarVsVarPlot_kwargs)

            # Reverse the change for the next plot
            VarVsVarPlot_kwargs["atlas_second_tag"] = unmodified_second_tag

    else:
        for flavour in flavour_list:
            unmodified_second_tag = VarVsVarPlot_kwargs.get("atlas_second_tag")

            # Modify the second tag
            if unmodified_second_tag is None:
                VarVsVarPlot_kwargs["atlas_second_tag"] = flavour.label

            else:
                VarVsVarPlot_kwargs["atlas_second_tag"] += f"\n{flavour.label}"

            # Init the plot
            var_plot_dict[flavour.name] = VarVsVarPlot(**VarVsVarPlot_kwargs)

            # Reverse the change for the next plot
            VarVsVarPlot_kwargs["atlas_second_tag"] = unmodified_second_tag

    # Init lists for legend handles
    file_handles = []
    flavour_handles = []
    trk_origin_handles = []

    # Iterate over the given files
    for file_counter, (file_key, file_value) in enumerate(files.items()):
        # Get the variable names for the given file
        jet_pt_variable = file_value.get("jet_pt_variable", "pt")
        flavour_label_variable = file_value.get(
            "flavour_label_variable",
            "HadronConeExclTruthLabelID",
        )
        track_truth_variable = file_value.get(
            "track_truth_variable",
            "ftagTruthOriginLabel",
        )

        # Get the name of the jets and tracks in the h5 files
        jets_name = file_value.get("jets_name", "jets")
        tracks_name = file_value.get("tracks_name", "tracks")

        # Get the pT bins for the given file
        pt_bins = file_value["pt_bins"]

        # Init the reader for the file
        reader = H5Reader(
            fname=file_value["filepath"],
            shuffle=False,
        )

        # Iterate over the flavour and load them from the file
        for flavour_counter, flavour in enumerate(flavour_list):
            # Get the iterator to correctly choose the plot to add to
            plot_iterator = file_key if all_flav_plot else flavour.name

            # Load the data
            data = reader.load(
                variables={
                    jets_name: [jet_pt_variable, flavour_label_variable],
                    tracks_name: [track_truth_variable],
                },
                num_jets=file_value.get("n_jets", None),
                cuts=flavour.cuts,
            )

            # Loop over the different track origins
            for trk_origin_counter, (trk_origin_key, track_origin_value) in enumerate(
                track_origin_dict.items()
            ):
                # Get the number of tracks from this specific origin
                n_trks_tmp = np.sum(
                    np.isin(
                        data[tracks_name][track_truth_variable],
                        track_origin_value,
                    ),
                    axis=1,
                )

                # Bin it in pT
                bin_indices = np.digitize(
                    data[jets_name][jet_pt_variable],
                    pt_bins,
                )

                # Calculate mean n_trks for the given origin in pT bins
                n_trks_means = np.array([
                    np.mean(n_trks_tmp[bin_indices == i]) for i in range(len(pt_bins) - 1)
                ])

                # Calculate the std error
                n_trks_std = [
                    np.sqrt(
                        np.sum(
                            (n_trks_tmp[bin_indices == i] - np.mean(n_trks_tmp[bin_indices == i]))
                            ** 2
                        )
                        / len(n_trks_tmp[bin_indices == i] - 1)
                    )
                    / np.sqrt(len(n_trks_tmp[bin_indices == i]))
                    for i in range(len(pt_bins) - 1)
                ]

                # Plot the curve
                var_plot_dict[plot_iterator].add(
                    VarVsVar(
                        x_var=(0.5 * (pt_bins[1:] + pt_bins[:-1])) / 1000,
                        y_var_mean=n_trks_means,
                        x_var_widths=(
                            (0.5 * (pt_bins[1:] + pt_bins[:-1])) / 1000 - pt_bins[:-1] / 1000
                        )
                        * 2,
                        y_var_std=n_trks_std,
                        plot_y_std=False,
                        linestyle=(
                            get_good_linestyles()[flavour_counter]
                            if all_flav_plot
                            else get_good_linestyles()[file_counter]
                        ),
                        marker="o",
                        markersize="4",
                        colour=get_good_colours()[trk_origin_counter],
                        label=None if all_flav_plot else file_value["process_label"],
                    )
                )

                # Append the track origin labels only once
                if flavour_counter == 0 and file_counter == 0:
                    trk_origin_handles.append(
                        mpl.lines.Line2D(
                            [],
                            [],
                            color=get_good_colours()[trk_origin_counter],
                            label=trk_origin_key,
                            linestyle=get_good_linestyles()[flavour_counter],
                            marker="o",
                            markersize="4",
                        )
                    )

            # Append the flavour labels only once per file
            if file_counter == 0:
                # Add the flavour to the legend
                flavour_handles.append(
                    mpl.lines.Line2D(
                        [],
                        [],
                        color="#000000",
                        label=flavour.label,
                        linestyle=get_good_linestyles()[flavour_counter],
                    )
                )

        # Add the process to the legend
        file_handles.append(
            mpl.lines.Line2D(
                [],
                [],
                color="#000000",
                label=file_value["process_label"],
                linestyle=get_good_linestyles()[file_counter],
            )
        )

    # Get the extra for the plot_name
    tmp_plot_name = "" if plot_name is None else plot_name + "_"

    # Iterate over all plots
    for iter_plot_name in var_plot_dict:
        # Draw the actual plot
        var_plot_dict[iter_plot_name].draw()

        # Remove initial legend
        var_plot_dict[iter_plot_name].axis_top.get_legend().remove()

        # Remove the puma legend and call adjust layout
        var_plot_dict[iter_plot_name].fig.tight_layout()

        # Add the flavour or the process legend to the plots
        iter_handles = flavour_handles if all_flav_plot else file_handles
        var_plot_dict[iter_plot_name].axis_top.add_artist(
            var_plot_dict[iter_plot_name].axis_top.legend(
                handles=iter_handles,
                labels=[handle.get_label() for handle in iter_handles],
                loc="upper right",
                ncol=1,
                frameon=False,
            )
        )

        # Add the track origin legend to the plots
        var_plot_dict[iter_plot_name].axis_top.add_artist(
            var_plot_dict[iter_plot_name].axis_top.legend(
                handles=trk_origin_handles,
                labels=[handle.get_label() for handle in trk_origin_handles],
                loc="upper center",
                ncol=2,
                frameon=False,
            )
        )

        # Safe the figure
        var_plot_dict[iter_plot_name].savefig(
            plot_name=os.path.join(
                plot_path,
                (
                    f"{tmp_plot_name}{iter_plot_name}_all_flavour.{plot_format}"
                    if all_flav_plot
                    else f"{tmp_plot_name}{iter_plot_name}_all_samples.{plot_format}"
                ),
            )
        )



"""Unit test script for the functions in utils.logging."""

from __future__ import annotations

import unittest

from puma.utils import logger, set_log_level
from puma.utils.logging import get_log_level

set_log_level(logger, "DEBUG")


class LogLevelTestCase(unittest.TestCase):
    """Test class for the puma logger."""

    def test_wrong_input(self):
        """Test scenario when a wrong log level is being provided."""
        with self.assertRaises(ValueError):
            get_log_level("TEST")



"""ROC curve functions."""

from __future__ import annotations

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from ftag import Flavours, Label

from puma.metrics import rej_err
from puma.plot_base import PlotBase, PlotLineObject
from puma.utils import get_good_colours, get_good_linestyles, logger


def can_hide(ax) -> bool:
    num_labels = 0
    ax_bbox = ax.get_window_extent()
    for label in ax.get_yticklabels():
        label_bbox = label.get_window_extent()
        if label.get_visible() and (label_bbox.y0 > ax_bbox.y0 and label_bbox.y1 < ax_bbox.y1):
            num_labels += 1
    return num_labels > 1


def adjust_ylabels(fig, axes, min_distance=1) -> None:
    """Adjust the y-axis labels to avoid overlap."""
    fig.canvas.draw()
    for ax in axes:
        labels = ax.get_yticklabels()
        ax_bbox = ax.get_window_extent()
        for label in labels:
            label_bbox = label.get_window_extent()
            # skip label if it is not visible
            if label_bbox.y1 < ax_bbox.y0 or label_bbox.y0 > ax_bbox.y1:
                continue
            # hide label if it is too close to the bottom
            if label_bbox.y0 - ax_bbox.y0 < min_distance and can_hide(ax):
                label.set_visible(False)
                continue
            # hide label if it is too close to the top
            if ax_bbox.y1 - label_bbox.y1 < min_distance and can_hide(ax):
                label.set_visible(False)
                continue


class Roc(PlotLineObject):
    """Represent a single ROC curve and allows to calculate ratio w.r.t other ROCs."""

    def __init__(
        self,
        sig_eff: np.ndarray,
        bkg_rej: np.ndarray,
        n_test: int | None = None,
        rej_class: str | Label = None,
        signal_class: str | None = None,
        key: str | None = None,
        ratio_group: str | None = None,
        **kwargs,
    ) -> None:
        """Initialise properties of roc curve object.

        Parameters
        ----------
        sig_eff : np.array
            Array of signal efficiencies
        bkg_rej : np.array
            Array of background rejection
        n_test : int
            Number of events used to calculate the background efficiencies,
            by default None
        signal_class : str
            Signal class, e.g. for b-tagging "bjets", by default None
        rej_class : str or Label
            Rejection class, e.g. for b-tagging anc charm rejection "cjets",
            by default None
        key : str
            Identifier for roc curve e.g. tagger, by default None
        ratio_group : str, optional
            Identifies the reference ROC group for ratio calculation, by default None
        **kwargs : kwargs
            Keyword arguments passed to `puma.PlotLineObject`

        Raises
        ------
        ValueError
            If `sig_eff` and `bkg_rej` have a different shape
        """
        super().__init__(**kwargs)
        if len(sig_eff) != len(bkg_rej):
            raise ValueError(
                f"The shape of `sig_eff` ({np.shape(sig_eff)}) and `bkg_rej` "
                f"({np.shape(bkg_rej)}) have to be identical."
            )
        self.sig_eff = sig_eff
        self.bkg_rej = bkg_rej
        self.n_test = None if n_test is None else int(n_test)
        self.signal_class = signal_class
        self.rej_class = Flavours[rej_class] if isinstance(rej_class, str) else rej_class
        self.key = key
        self.ratio_group = ratio_group if ratio_group else str(rej_class)

    def binomial_error(self, norm: bool = False, n_test: int | None = None) -> np.ndarray:
        """Calculate binomial error of roc curve.

        Parameters
        ----------
        norm : bool
            If True calulate relative error, by default False
        n_test : int
            Number of events used to calculate the background efficiencies,
            by default None

        Returns
        -------
        numpy.array
            Binomial error

        Raises
        ------
        ValueError
            If no `n_test` was provided
        """
        if n_test is None:
            n_test = self.n_test
        if n_test is None:
            raise ValueError("No `n_test` provided, cannot calculate binomial error!")
        return rej_err(self.bkg_rej[self.non_zero_mask], n_test, norm=norm)

    def divide(self, roc_comp, inverse: bool = False):
        """Calculate ratio between the roc curve and another roc.

        Parameters
        ----------
        roc_comp : roc class
            Second roc curve to calculate ratio with
        inverse : bool
            If False the ratio is calculated `this_roc / roc_comp`,
            if True the inverse is calculated

        Returns
        -------
        np.array
            Signal efficiency used for the ratio calculation which is the overlapping
            interval of the two roc curves
        np.array
            Ratio
        np.array or None
            Ratio_err if `n_test` was provided to class
        """
        if not np.array_equal(self.sig_eff, roc_comp.sig_eff):
            raise ValueError("Signal efficiencies of the two ROCs do not match.")

        ratio = self.bkg_rej / roc_comp.bkg_rej
        if inverse:
            ratio = 1 / ratio

        ratio_err = self.binomial_error(norm=True) * ratio if self.n_test else None
        return self.sig_eff, ratio, ratio_err

    @property
    def non_zero_mask(self):
        """Masking points where rejection is 0 and no signal efficiency change present.

        Returns
        -------
        numpy.array
            Masked indices
        """
        # Mask the points where there was no change in the signal eff
        delta_x = np.concatenate((np.ones(1), np.diff(self.sig_eff)))

        # Also mask the rejections that are 0
        nonzero = (self.bkg_rej != 0) & (delta_x > 0)
        if self.xmin is not None:
            nonzero = nonzero & (self.sig_eff >= self.xmin)
        if self.xmax is not None:
            nonzero = nonzero & (self.sig_eff <= self.xmax)
        return nonzero

    @property
    def non_zero(self):
        """Abstraction of `non_zero_mask`.

        Returns
        -------
        numpy.array
            Masked signal efficiency
        numpy.array
            Masked background rejection
        """
        return self.sig_eff[self.non_zero_mask], self.bkg_rej[self.non_zero_mask]


class RocPlot(PlotBase):
    """ROC plot class."""

    def __init__(self, grid: bool = True, **kwargs) -> None:
        """ROC plot properties.

        Parameters
        ----------
        grid : bool, optional
            Set the grid for the plots.
        **kwargs : kwargs
            Keyword arguments from `puma.PlotObject`
        """
        super().__init__(grid=grid, **kwargs)
        self.test = ""
        self.rocs = {}
        self.roc_ratios = {}
        self.rej_axes = {}
        self.rej_class_ls = {}
        self.label_colours = {}
        self.leg_rej_labels = {}
        self.reference_roc = None
        self.initialise_figure()
        self.fig.get_layout_engine().set(h_pad=0, hspace=0)
        self.eff_min, self.eff_max = (1, 0)
        self.default_linestyles = get_good_linestyles()
        self.legend_flavs = None
        self.rej_leg_loc = "ratio" if kwargs["n_ratio_panels"] > 0 else "lower left"

    def add_roc(
        self,
        roc_curve: object,
        key: str | None = None,
        reference: bool = False,
    ):
        """Adding puma.Roc object to figure.

        Parameters
        ----------
        roc_curve : puma.Roc
            ROC curve
        key : str, optional
            Unique identifier for roc_curve, by default None
        reference : bool, optional
            If roc is used as reference for ratio calculation, by default False

        Raises
        ------
        KeyError
            If unique identifier key is used twice
        """
        if key is None:
            key = len(self.rocs) + 1
        if key in self.rocs:
            raise KeyError(f"Duplicated key {key} already used for roc unique identifier.")

        self.rocs[key] = roc_curve
        # set linestyle
        if roc_curve.rej_class not in self.rej_class_ls:
            self.rej_class_ls[roc_curve.rej_class] = (
                self.default_linestyles[len(self.rej_class_ls)]
                if roc_curve.linestyle is None
                else roc_curve.linestyle
            )
        elif (
            roc_curve.linestyle != self.rej_class_ls[roc_curve.rej_class]
            and roc_curve.linestyle is not None
        ):
            logger.warning(
                "You specified a different linestyle for the same rejection class "
                "%s. Will keep the linestyle defined first.",
                roc_curve.rej_class,
            )
        if roc_curve.linestyle is None:
            roc_curve.linestyle = self.rej_class_ls[roc_curve.rej_class]

        # set colours
        if roc_curve.label not in self.label_colours:
            self.label_colours[roc_curve.label] = (
                get_good_colours()[len(self.label_colours)]
                if roc_curve.colour is None
                else roc_curve.colour
            )
        elif (
            roc_curve.colour != self.label_colours[roc_curve.label] and roc_curve.colour is not None
        ):
            logger.warning(
                "You specified a different colour for the same label"
                " %s. This will lead to a mismatch in the line colours"
                " and the legend.",
                roc_curve.label,
            )
        if roc_curve.colour is None:
            roc_curve.colour = self.label_colours[roc_curve.label]

        if reference:
            logger.debug("Setting roc %s as reference for %s.", key, roc_curve.rej_class)
            self.set_roc_reference(key, roc_curve.rej_class, roc_curve.ratio_group)
            self.reference_label = roc_curve.label

    def set_roc_reference(
        self,
        key: str,
        rej_class: Label,
        ratio_group: str | None = None,
    ):
        """Setting the reference roc curves used in the ratios.

        Parameters
        ----------
        key : str
            Unique identifier of roc object
        rej_class : str
            Rejection class encoded in roc curve
        ratio_group : str
            Ratio group this roc is reference for, by default None

        Raises
        ------
        ValueError
            If more rejection classes are set than actual ratio panels available.
        """
        if self.reference_roc is None:
            self.reference_roc = {}
            self.reference_roc[rej_class] = {ratio_group: key}
        elif rej_class not in self.reference_roc:
            if len(self.reference_roc) >= self.n_ratio_panels:
                raise ValueError(
                    "You cannot set more rejection classes than available ratio panels."
                )
            self.reference_roc[rej_class] = {ratio_group: key}
        else:
            if self.reference_roc[rej_class].get(ratio_group):
                logger.warning(
                    "You specified a second roc curve %s as reference for ratio. "
                    "Using it as new reference instead of %s.",
                    key,
                    self.reference_roc[rej_class][ratio_group],
                )
            self.reference_roc[rej_class][ratio_group] = key

    def set_ratio_class(self, ratio_panel: int, rej_class: str | Label):
        """Associate the rejection class to a ratio panel adn set the legend label.

        Parameters
        ----------
        ratio_panel : int
            Ratio panel either 1 or 2
        rej_class : Labels
            Rejeciton class associated to that panel

        Raises
        ------
        ValueError
            if requested ratio panels and given ratio_panel do not match.
        """
        rej_class = Flavours[rej_class] if isinstance(rej_class, str) else rej_class
        self.rej_axes[rej_class] = self.ratio_axes[ratio_panel - 1]
        label = rej_class.label.replace("jets", "jet")
        self.set_ratio_label(ratio_panel, f"{label} ratio")
        self.leg_rej_labels[rej_class] = rej_class.label

    def add_ratios(self):
        """Calculating ratios.

        Raises
        ------
        ValueError
            If number of reference rocs and ratio panels don't match
        ValueError
            If no ratio classes are set
        """
        if self.reference_roc and len(self.reference_roc) != self.n_ratio_panels:
            raise ValueError(
                f"{len(self.reference_roc)} reference rocs defined but requested "
                f"{self.n_ratio_panels} ratio panels."
            )
        if len(self.rej_axes) != self.n_ratio_panels:
            raise ValueError("Ratio classes not set, set them first with `set_ratio_class`.")

        for rej_class, axis in self.rej_axes.items():
            self.plot_ratios(axis=axis, rej_class=rej_class)

    def get_xlim_auto(self):
        """Returns min and max efficiency values.

        Returns
        -------
        float
            Min and max efficiency values
        """
        for elem in self.rocs.values():
            self.eff_min = min(np.min(elem.sig_eff), self.eff_min)
            self.eff_max = max(np.max(elem.sig_eff), self.eff_min)

        return self.eff_min, self.eff_max

    def plot_ratios(self, axis: plt.axis, rej_class: str):
        """Plotting ratio curves.

        Parameters
        ----------
        axis : plt.axis
            matplotlib axis object
        rej_class : str
            Rejection class
        """
        for key, elem in self.rocs.items():
            if elem.rej_class != rej_class:
                continue
            sig_eff, ratio, ratio_err = elem.divide(
                self.rocs[self.reference_roc[rej_class][elem.ratio_group]]
            )

            self.roc_ratios[key] = (sig_eff, ratio, ratio_err)
            axis.plot(
                sig_eff,
                ratio,
                color=elem.colour,
                linestyle=elem.linestyle,
                linewidth=2.0,
            )
            if ratio_err is not None:
                axis.fill_between(
                    sig_eff,
                    ratio - ratio_err,
                    ratio + ratio_err,
                    color=elem.colour,
                    alpha=0.25,
                    edgecolor="none",
                    zorder=1,
                )

    def make_split_legend(self, handles):
        """Draw legend for the case of 2 ratios, splitting up legend into models and
        rejection class.

        Parameters
        ----------
        handles : list
            List of Line2D objects to extract info for legend

        Raises
        ------
        ValueError
            If not 2 ratios requested
        """
        if self.n_ratio_panels < 2:
            raise ValueError("For a split legend you need 2 ratio panels.")

        if self.rej_leg_loc == "ratio":
            for rej_class, axis in self.rej_axes.items():
                legend_line = mpl.lines.Line2D(
                    [],
                    [],
                    color="k",
                    label=self.leg_rej_labels[rej_class],
                    linestyle=self.rej_class_ls[rej_class],
                )
                axis.legend(
                    handles=[legend_line],
                    labels=[legend_line.get_label()],
                    loc="upper right",
                    fontsize=self.leg_fontsize,
                )

        else:
            line_list_rej = [
                mpl.lines.Line2D(
                    [],
                    [],
                    color="k",
                    label=self.leg_rej_labels[rej_class],
                    linestyle=self.rej_class_ls[rej_class],
                )
                for rej_class in self.rej_axes
            ]

            self.legend_flavs = self.axis_top.legend(
                handles=line_list_rej,
                labels=[handle.get_label() for handle in line_list_rej],
                loc=self.rej_leg_loc,
                fontsize=self.leg_fontsize,
                ncol=self.leg_ncol,
            )

            # Add the second legend to plot
            self.axis_top.add_artist(self.legend_flavs)

        # Get the labels for the legends
        labels_list = []
        lines_list = []

        for line in handles:
            if line.get_label() not in labels_list:
                labels_list.append(line.get_label())
                lines_list.append(line)

        # Define the legend
        self.axis_top.legend(
            handles=lines_list,
            labels=labels_list,
            loc=self.leg_loc,
            fontsize=self.leg_fontsize,
            ncol=self.leg_ncol,
        )

    def draw(
        self,
        labelpad: int | None = None,
    ):
        """Draw plotting.

        Parameters
        ----------
        labelpad : int, optional
            Spacing in points from the axes bounding box including
            ticks and tick labels, by default None
        """
        plt_handles = self.plot_roc()
        xmin, xmax = self.get_xlim_auto()

        self.set_xlim(
            xmin if self.xmin is None else self.xmin,
            xmax if self.xmax is None else self.xmax,
        )
        if self.n_ratio_panels > 0:
            self.add_ratios()
        self.set_title()
        self.set_log()
        self.set_y_lim()
        self.set_xlabel()
        self.set_ylabel(self.axis_top)

        # set ylabel for ratio panels
        if self.n_ratio_panels > 0:
            self.set_ylabel(
                list(self.rej_axes.values())[-1],
                f"Ratio to {self.reference_label}",
                align="left",
                labelpad=labelpad,
            )

        if self.n_ratio_panels < 2:
            self.make_legend(plt_handles, ax_mpl=self.axis_top)
        else:
            if not self.leg_rej_labels:
                for rej_class in self.rej_axes:
                    self.leg_rej_labels[rej_class] = rej_class
            self.make_split_legend(handles=plt_handles)

        self.plotting_done = True
        if self.apply_atlas_style is True:
            self.atlasify()
            # atlasify can only handle one legend. Therefore, we remove the frame of
            # the second legend by hand
            if self.legend_flavs is not None:
                self.legend_flavs.set_frame_on(False)

        adjust_ylabels(self.fig, self.rej_axes.values())

    def plot_roc(self, **kwargs) -> mpl.lines.Line2D:
        """Plotting roc curves.

        Parameters
        ----------
        **kwargs: kwargs
            Keyword arguments passed to plt.axis.plot

        Returns
        -------
        Line2D
            matplotlib Line2D object
        """
        plt_handles = []
        for key, elem in self.rocs.items():
            plt_handles = plt_handles + self.axis_top.plot(
                elem.sig_eff[elem.non_zero_mask],
                elem.bkg_rej[elem.non_zero_mask],
                linestyle=elem.linestyle,
                linewidth=2,
                color=elem.colour,
                label=elem.label if elem is not None else key,
                zorder=2,
                **kwargs,
            )
            if elem.n_test is not None:
                # if uncertainties are available for roc plotting their uncertainty as
                # a band around the roc itself
                rej_band_down = (
                    elem.bkg_rej[elem.non_zero_mask] - elem.binomial_error()[elem.non_zero_mask]
                )
                rej_band_up = (
                    elem.bkg_rej[elem.non_zero_mask] + elem.binomial_error()[elem.non_zero_mask]
                )
                self.axis_top.fill_between(
                    elem.sig_eff[elem.non_zero_mask],
                    rej_band_down,
                    rej_band_up,
                    color=elem.colour,
                    alpha=0.25,
                    edgecolor="none",
                    zorder=2,
                )
        return plt_handles



"""Unit test script for the functions in utils/generate.py."""

from __future__ import annotations

import unittest

import numpy as np
from numpy.lib.recfunctions import structured_to_unstructured as s2u

from puma.utils import logger, set_log_level
from puma.utils.generate import (
    get_dummy_2_taggers,
    get_dummy_multiclass_scores,
    get_dummy_tagger_aux,
)

set_log_level(logger, "DEBUG")


class GetDummyMulticlassScoresTestCase(unittest.TestCase):
    """Test case for get_dummy_multiclass_scores function."""

    def test_size(self):
        """Check that correct size is returned."""
        output, labels = get_dummy_multiclass_scores(size=10)
        # we expect here 9 entries, since 10 is not dividable by 3, and the function
        # returns 3 classes with the same amount of stats per class
        with self.subTest("output length"):
            self.assertEqual(len(output), 9)
        with self.subTest("label length"):
            self.assertEqual(len(labels), 9)

    def test_range(self):
        """Check that correct range of output is returned."""
        output, _ = get_dummy_multiclass_scores()
        with self.subTest("max val"):
            self.assertLessEqual(np.max(s2u(output)), 1)
        with self.subTest("min val"):
            self.assertGreaterEqual(np.min(s2u(output)), 0)


class GetDummy2TaggersTestCase(unittest.TestCase):
    """Test case for get_dummy_2_taggers function."""

    def test_size(self):
        """Check that correct size is returned."""
        df_gen = get_dummy_2_taggers(size=10)
        # we expect here 9 entries, since 10 is not dividable by 3, and the function
        # returns 3 classes with the same amount of stats per class
        self.assertEqual(len(df_gen), 9)

    def test_columns(self):
        """Check correct amount of columns."""
        df_gen = get_dummy_2_taggers()
        self.assertEqual(len(df_gen.dtype.names), 9)

    def test_columns_pt(self):
        """Check correct amount of columns when using pt as well."""
        df_gen = get_dummy_2_taggers(add_pt=True)
        self.assertEqual(len(df_gen.dtype.names), 10)


class GetDummyTaggerAuxTestCase(unittest.TestCase):
    """Test case for get_dummy_tagger_aux function."""

    def test_size(self):
        """Check that correct size is returned."""
        _, file = get_dummy_tagger_aux(size=10)
        # we expect here 9 entries, since 10 is not dividable by 3, and the function
        # returns 3 classes with the same amount of stats per class
        self.assertEqual(file["jets"].shape[0], 9)

    def test_ntracks(self):
        """Check correct amount of tracks."""
        _, file = get_dummy_tagger_aux(n_tracks=10)
        self.assertEqual(file["tracks"].shape[1], 10)

    def test_columns(self):
        """Check correct amount of columns for jets and tracks."""
        _, file = get_dummy_tagger_aux()
        self.assertEqual(len(file["jets"][0]), 8)
        self.assertEqual(len(file["tracks"][0][0]), 8)



"""Support functions for vertexing performance in flavour tagging."""

from __future__ import annotations

import numpy as np


def clean_indices(vertex_ids, condition, mode="remove"):
    """
    Vertex index cleaner that modifies vertex indices that fulfill
    a specified condition. The mode can be set to either "remove"
    or "merge". In the "remove" mode, the vertex indices are set to
    be ignored by the matching algorithm while in the "merge" mode
    the vertex indices are merged into a single vertex.

    Parameters
    ----------
    vertex_ids: np.ndarray
        Array containing vertex IDs for each track.
    condition: np.ndarray
        Boolean array of shape (n_jets, n_tracks) containing the condition
        to be applied.
    mode: str
        Mode to apply to indices that meet the specified condition.
        Options are "remove" and "merge".

    Returns
    -------
    vertex_ids: np.ndarray
        Array containing vertex IDs for each track.

    Raises
    ------
    ValueError
        If the mode is not recognized.
    """
    if mode == "remove":
        vertex_ids[condition] = -99
    elif mode == "merge":
        if len(set(vertex_ids[condition])) > 1:
            vertex_ids[condition] = np.max(vertex_ids) + 1
    else:
        raise ValueError(f"Mode {mode} not recognized.")

    return vertex_ids


def build_vertices(vertex_ids):
    """
    Vertex builder that outputs an array of vertex associations
    from vertex ids derived in athena or salt for a single jet.
    Negative indices and one-track vertices are ignored by
    default.

    Parameters
    ----------
    vertex_ids: np.ndarray
        Array containing vertex IDs for each track.

    Returns
    -------
    vertices: np.ndarray
        Boolean array of shape (n_vertices, n_tracks) containing track-vertex
        associations for each vertex. Each track is associated with at most
        one vertex.
    """
    unique_ids, unique_counts = np.unique(vertex_ids, return_counts=True)
    unique_ids = unique_ids[unique_counts > 1]  # remove vertices with only one track
    unique_ids = unique_ids[unique_ids >= 0]  # remove tracks with negative indices

    vertices = np.tile(vertex_ids, (unique_ids.size, 1))
    comparison_ids = np.tile(unique_ids, (vertex_ids.size, 1)).T
    return vertices == comparison_ids


def associate_vertices(test_vertices, ref_vertices, eff_req, purity_req):
    """
    Vertex associator that maps two collections of vertices onto
    each other 1-to-1 based on the highest overlap of track indices
    for a single jet. Percentage of overlapping and non-overlapping
    tracks relative to vertex size are used as tiebreakers. The
    matching is performed in a greedy fashion and decisions are not
    revisited.

    Parameters
    ----------
    test_vertices: np.ndarray
        Boolean array of shape (n_test_vertices, n_tracks) containing track-vertex
        associations for vertex collection to be tested (reco).
    ref_vertices: np.ndarray
        Boolean array of shape (n_ref_vertices, n_tracks) containing track-vertex
        associations for vertex collection to use as reference (truth).
    eff_req: float, optional
        Minimum required efficiency for vertex matching.
    purity_req: float, optional
        Minimum required purity for vertex matching.

    Returns
    -------
    associations: np.ndarray
        Boolean matrix of vertex associations with shape (n_test_vertices,
        n_ref_vertices).
    common_tracks: np.ndarray
        Matrix containing number of common tracks shared by each vertex pairing.
    """
    ref_sizes = ref_vertices.sum(axis=1)
    test_sizes = test_vertices.sum(axis=1)
    n_ref = ref_sizes.size
    n_test = test_sizes.size

    common_tracks = np.dot(test_vertices.astype(int), ref_vertices.astype(int).T)
    inv_ref_size = 1.0 / np.tile(ref_sizes, (n_test, 1))
    inv_test_size = 1.0 / np.tile(test_sizes, (n_ref, 1)).T

    pair_index = np.arange(n_ref * n_test, 0, -1).reshape(
        n_test, n_ref
    )  # unique number for each vertex pairing

    # calculate vertex associations based on maximum number of shared tracks
    # with ties broken by preferring the smallest true vertex (maximum purity score)
    # followed by the smallest reco vertex (minimum false association); if there are
    # two equally good pairings for a vertex, the first in the list is chosen
    associations = np.ones_like(common_tracks, dtype=bool)
    for metric in [common_tracks, inv_ref_size, inv_test_size, pair_index]:
        metric[np.logical_not(associations)] = -1
        col_max = np.tile(np.amax(metric, axis=0), (n_test, 1))
        row_max = np.tile(np.amax(metric, axis=1), (n_ref, 1)).T
        associations = np.logical_and(metric == col_max, metric == row_max)
        associations[metric == -1] = False  # remove pairs that were already excluded
    associations[common_tracks == 0] = False  # remove leftover pairs with zero matches

    # enforce purity and efficiency requirements
    eff_cut = common_tracks * inv_ref_size >= eff_req
    purity_cut = common_tracks * inv_test_size >= purity_req
    associations = np.logical_and.reduce((associations, eff_cut, purity_cut))

    return associations, common_tracks


def calculate_vertex_metrics(
    test_indices,
    ref_indices,
    max_vertices=20,
    eff_req=0.65,
    purity_req=0.5,
):
    """
    Vertex metric calculator that outputs a set of metrics useful for evaluating
    vertexing performance for each jet.

    Parameters
    ----------
    test_indices: np.ndarray
        Boolean array of shape (n_jets, n_tracks) containing vertex indices to be
        tested (reco).
    ref_indices: np.ndarray
        Boolean array of shape (n_jets, n_tracks) containing vertex indices to use
        as reference (truth).
    max_vertices: int, optional
        Maximum number of matched vertices to write out, by default 20.
    eff_req: float, optional
        Minimum required efficiency for vertex matching, by default 0.65.
    purity_req: float, optional
        Minimum required purity for vertex matching, by default 0.5.

    Returns
    -------
    metrics: dict
        Dictionary containing the following metrics:
            n_match: np.ndarray
                Array of shape (n_jets) containing the number of matched vertices per
                jet.
            n_test: np.ndarray
                Array of shape (n_jets) containing the number of reco vetices per jet.
            n_ref: np.ndarray
                Array of shape (n_jets) containing the number of truth vertices per jet.
            track_overlap: np.ndarray
                Array of shape (n_jets, max_vertices) containing the number of
                overlapping tracks between each matched vertex pair.
            test_vertex_size: np.ndarray
                Array of shape (n_jets, max_vertices) containing the number of tracks
                in each matched reco vertex.
            ref_vertex_size: np.ndarray
                Array of shape (n_jets, max_vertices) containing the number of tracks
                in each matched truth vertex.
    """
    assert (
        ref_indices.shape == test_indices.shape
    ), "Truth and reco vertex arrays must have the same shape."
    n_jets = ref_indices.shape[0]

    metrics = {}
    metrics["n_match"] = np.zeros(n_jets, dtype=int)
    metrics["n_test"] = np.zeros(n_jets, dtype=int)
    metrics["n_ref"] = np.zeros(n_jets, dtype=int)
    metrics["track_overlap"] = np.full((n_jets, max_vertices), -1)
    metrics["test_vertex_size"] = np.full((n_jets, max_vertices), -1)
    metrics["ref_vertex_size"] = np.full((n_jets, max_vertices), -1)

    for i in range(n_jets):
        ref_vertices = build_vertices(ref_indices[i])
        test_vertices = build_vertices(test_indices[i])

        # handle edge cases
        if not ref_vertices.any() and not test_vertices.any():
            continue
        if not ref_vertices.any():
            metrics["n_test"][i] = test_vertices.shape[0]
            continue
        if not test_vertices.any():
            metrics["n_ref"][i] = ref_vertices.shape[0]
            continue

        associations, common_tracks = associate_vertices(
            test_vertices,
            ref_vertices,
            eff_req=eff_req,
            purity_req=purity_req,
        )

        # write out vertexing efficiency metrics
        metrics["n_match"][i] = np.sum(associations)
        metrics["n_ref"][i] = ref_vertices.shape[0]
        metrics["n_test"][i] = test_vertices.shape[0]

        # only save purity metrics for requested number of vertices
        max_index = min(metrics["n_match"][i], max_vertices)

        # write out vertexing purity metrics
        metrics["track_overlap"][i, :max_index] = common_tracks[associations][:max_index]
        metrics["test_vertex_size"][i, :max_index] = test_vertices[
            associations.sum(axis=1).astype(bool)
        ].sum(axis=1)[:max_index]
        metrics["ref_vertex_size"][i, :max_index] = ref_vertices[
            associations.sum(axis=0).astype(bool)
        ].sum(axis=1)[:max_index]

    return metrics


def clean_truth_vertices(truth_vertices, truth_track_origin, incl_vertexing=False):
    """
    Clean truth vertices for each track in a single jet. This function removes
    all truth vertices that are not entirely from HF. If inclusive vertexing
    is enabled, it also merges remaining vertices into a single vertex.

    Parameters
    ----------
    truth_vertices: np.ndarray
        Array containing truth vertex indices for each track in a jet.
    truth_track_origin: np.ndarray
        Array containing truth track origin labels for each track in a jet.
    incl_vertexing: bool, optional
        Whether to merge all vertex indices, by default False.

    Returns
    -------
    truth_vertices: np.ndarray
        Array containing cleaned truth vertex indices for each track in a jet.
    """
    # remove vertices that aren't purely HF
    removal_indices = np.unique(truth_vertices[np.isin(truth_track_origin, [3, 4, 5], invert=True)])
    truth_vertices = clean_indices(
        truth_vertices,
        np.isin(truth_vertices, removal_indices),
        mode="remove",
    )

    # merge truth vertices from HF for inclusive performance
    if incl_vertexing:
        truth_vertices = clean_indices(
            truth_vertices,
            truth_vertices > 0,
            mode="merge",
        )

    return truth_vertices


def clean_reco_vertices(reco_vertices, reco_track_origin=None, incl_vertexing=False):
    """
    Clean reconstructed vertices for each track in a single jet. This function
    removes the vertex with the most reco PV tracks if track origin classification
    information is available. If inclusive vertexing is enabled, it also merges all
    vertices with > 0 tracks from HF and removes all others for taggers with track
    origin classification while merging all vertices for taggers without it.

    Parameters
    ----------
    reco_vertices: np.ndarray
        Array containing reco vertex indices for each track in a jet.
    reco_track_origin: np.ndarray, optional
        Array containing reco track origin labels for each track in a jet.
    incl_vertexing: bool, optional
        Whether to merge all vertex indices, by default False.

    Returns
    -------
    reco_vertices: np.ndarray
        Array containing cleaned reco vertex indices for each track in a jet.
    """
    # elaborate cleaning if track origin predictions are available
    if reco_track_origin is not None:
        # remove vertex with most reco PV tracks
        pv_candidate_indices, pv_candidate_counts = np.unique(
            reco_vertices[reco_track_origin == 2], return_counts=True
        )
        if pv_candidate_indices.size > 0:
            pv_index = pv_candidate_indices[np.argmax(pv_candidate_counts)]
            reco_vertices = clean_indices(
                reco_vertices,
                reco_vertices == pv_index,
                mode="remove",
            )

        # remove vertices with no tracks from HF
        hf_vertex_indices = np.unique(reco_vertices[np.isin(reco_track_origin, [3, 4, 5])])
        reco_vertices = clean_indices(
            reco_vertices,
            np.isin(reco_vertices, hf_vertex_indices, invert=True),
            mode="remove",
        )

        # merge remaining vertices for inclusive performance
        if incl_vertexing:
            reco_vertices = clean_indices(
                reco_vertices,
                reco_vertices >= 0,
                mode="merge",
            )

    # merge all reco vertices if track origin predictions are not available
    elif incl_vertexing:
        reco_vertices = clean_indices(
            reco_vertices,
            reco_vertices >= 0,
            mode="merge",
        )

    return reco_vertices



"""Module for usefule tools in puma."""

# flake8: noqa

import numpy as np
import pandas as pd
from palettable.colorbrewer.qualitative import Dark2_8

from puma.utils.aux import get_aux_labels
from puma.utils.generate import (
    get_dummy_2_taggers,
    get_dummy_tagger_aux,
    get_dummy_multiclass_scores,
)
from puma.utils.logging import logger, set_log_level


def set_xaxis_ticklabels_invisible(ax):
    """Helper function to set the ticklabels of the xaxis invisible

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        Axis you want to modify
    """

    for label in ax.get_xticklabels():
        label.set_visible(False)


def get_good_pie_colours(colour_scheme=None):
    """Helper function to get good colours for a pie chart. You can
    choose between a specific colour scheme or use the default colours
    for a pie chart

    Parameters
    ----------
    colour_scheme : string, optional
        colour scheme for the pie chart. Can be None to use default colours
        or blue, red, green or yellow to use a specific colour scheme

    Returns
    -------
    list
        returns a list of colours in the specified colour scheme

    Raises
    ------
    KeyError
        If colour_scheme is not in ["blue", "red", "green", "yellow", None]
    """
    if colour_scheme is None:
        return [
            "#3f90da",
            "#ffa90e",
            "#bd1f01",
            "#94a4a2",
            "#832db6",
            "#a96b59",
            "#e76300",
            "#b9ac70",
            "#717581",
            "#92dadd",
        ]
    if colour_scheme == "red":
        return [
            "#750000",
            "#A30000",
            "#D10000",
            "#FF0000",
            "#FF2E2E",
            "#FF5C5C",
            "#FF8A8A",
        ]
    if colour_scheme == "blue":
        return [
            "#000075",
            "#0000A3",
            "#0000D1",
            "#0000FF",
            "#2E2EFF",
            "#5C5CFF",
            "#8A8AFF",
        ]
    if colour_scheme == "green":
        return [
            "#1E5631",
            "#A4DE02",
            "#76BA1B",
            "#4C9A2A",
            "#ACDF87",
            "#68BB59",
            "#CEF3CE",
        ]
    if colour_scheme == "yellow":
        return [
            "#755800",
            "#A37A00",
            "#D19D00",
            "#FFBF00",
            "#FFCB2E",
            "#FFD65C",
            "#FFE28A",
        ]
    raise KeyError(
        f"Given colour scheme is {colour_scheme} but it has to be blue, red, green,"
        " yellow or None"
    )


def get_good_colours(colour_scheme: str | None = None):
    """List of colours adequate for plotting

    Parameters
    ----------
    colour_scheme : string, optional
        colour scheme for line plots, by default None

    Returns
    -------
    list
        list with colours
    """

    # If no colour scheme is selected, return colour-blind friendly colours
    # See https://arxiv.org/pdf/2107.02270 Page 15 (10 colours)
    if colour_scheme is None:
        return [
            "#3f90da",
            "#ffa90e",
            "#bd1f01",
            "#94a4a2",
            "#832db6",
            "#a96b59",
            "#e76300",
            "#b9ac70",
            "#717581",
            "#92dadd",
        ] + Dark2_8.mpl_colors

    elif colour_scheme == "Dark2_8":
        return Dark2_8.mpl_colors


def get_good_markers():
    """List of markers adequate for plotting

    Returns
    -------
    list
        list with markers
    """
    return [
        "o",  # Circle
        "x",  # x
        "v",  # Triangle down
        "^",  # Triangle up
        "D",  # Diamond
        "p",  # Pentagon
        "s",  # Square
    ]


def get_good_linestyles(names=None):
    """Returns a list of good linestyles

    Parameters
    ----------
    names : list or str, optional
        List or string of the name(s) of the linestyle(s) you want to retrieve, e.g.
        "densely dotted" or ["solid", "dashdot", "densely dashed"], by default None

    Returns
    -------
    list
        List of good linestyles. Either the specified selection or the whole list in
        the predefined order.

    Raises
    ------
    ValueError
        If `names` is not a str or list.
    """
    linestyle_tuples = {
        "solid": "solid",
        "densely dashdotted": (0, (3, 1, 1, 1)),
        "densely dashed": (0, (5, 2)),
        "densely dotted": (0, (1, 1)),
        "densely dashdotdotted": (0, (3, 1, 1, 1, 1, 1)),
        "dashed": (0, (5, 5)),
        "dotted": (0, (2, 2)),
        "dashdot": "dashdot",
        "loosely dashed": (0, (5, 10)),
        "loosely dotted": (0, (1, 10)),
        "loosely dashdotted": (0, (3, 10, 1, 10)),
        "loosely dashdotdotted": (0, (3, 10, 1, 10, 1, 10)),
        "dashdotted": (0, (3, 5, 1, 5)),
        "dashdotdotted": (0, (3, 5, 1, 5, 1, 5)),
    }

    default_order = [
        "solid",
        "densely dashdotted",
        "densely dotted",
        "densely dashed",
        "densely dashdotdotted",
        "dashed",
        "dotted",
        "dashdot",
        "loosely dashed",
        "loosely dotted",
        "loosely dashdotted",
        "loosely dashdotdotted",
        "dashdotted",
        "dashdotdotted",
    ]
    if names is None:
        names = default_order
    elif isinstance(names, str):
        return linestyle_tuples[names]
    elif not isinstance(names, list):
        raise ValueError("Invalid type of `names`, has to be a list of strings or a sting.")
    return [linestyle_tuples[name] for name in names]



from __future__ import annotations

import unittest

import numpy as np

from puma.utils import logger, set_log_level
from puma.utils.confusion_matrix import confusion_matrix

set_log_level(logger, "DEBUG")


class ConfusionMatrixTestCase(unittest.TestCase):
    def test_confusion_matrix_unweighted(self):
        targets = np.array([2, 0, 2, 2, 0, 1])
        predictions = np.array([0, 0, 2, 2, 0, 2])
        cm = confusion_matrix(targets, predictions, normalize="all")
        expected_cm = np.array([
            [0.33333333, 0.0, 0.0],
            [0.0, 0.0, 0.16666667],
            [0.16666667, 0.0, 0.33333333],
        ])
        np.testing.assert_array_almost_equal(expected_cm, cm)

    def test_confusion_matrix_unweighted_colnorm(self):
        targets = np.array([2, 0, 2, 2, 0, 1])
        predictions = np.array([0, 0, 2, 2, 0, 1])
        cm = confusion_matrix(targets, predictions, normalize="colnorm")
        expected_cm = np.array([[0.66666667, 0.0, 0.0], [0.0, 1.0, 0.0], [0.33333333, 0.0, 1.0]])
        np.testing.assert_array_almost_equal(expected_cm, cm)

    def test_confusion_matrix_unweighted_rownorm(self):
        targets = np.array([2, 0, 2, 2, 0, 1])
        predictions = np.array([0, 0, 2, 2, 0, 1])
        cm = confusion_matrix(targets, predictions, normalize="rownorm")
        expected_cm = np.array([
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
            [0.3333333333333333, 0.0, 0.6666666666666666],
        ])
        np.testing.assert_array_almost_equal(expected_cm, cm)

    def test_confusion_matrix_unweighted_normNone(self):
        targets = np.array([2, 0, 2, 2, 0, 1])
        predictions = np.array([0, 0, 2, 2, 0, 1])
        cm = confusion_matrix(targets, predictions, normalize=None)
        expected_cm = np.array([[2.0, 0.0, 0.0], [0.0, 1.0, 0.0], [1.0, 0.0, 2.0]])
        np.testing.assert_array_almost_equal(expected_cm, cm)

    def test_confusion_matrix_weighted(self):
        targets = np.array([2, 0, 2, 2, 0, 1])
        predictions = np.array([0, 0, 2, 2, 0, 2])
        weights = np.array([1, 0.5, 0.5, 1, 0.2, 1])
        cm = confusion_matrix(targets, predictions, sample_weights=weights, normalize="all")
        expected_cm = np.array([
            [0.16666667, 0.0, 0.0],
            [0.0, 0.0, 0.23809524],
            [0.23809524, 0.0, 0.35714286],
        ])
        np.testing.assert_array_almost_equal(expected_cm, cm)



"""Unit test script for the functions in utils/__init__.py."""

from __future__ import annotations

import os
import shutil  # noqa: F401
import tempfile
import unittest

import numpy as np
from matplotlib.testing.compare import compare_images

from puma import Line2D, Line2DPlot
from puma.utils import get_good_linestyles, logger, set_log_level

set_log_level(logger, "DEBUG")


class LinestylesTestCase(unittest.TestCase):
    """Test class for the linestyle management of puma."""

    def setUp(self):
        # Set up directories for comparison plots
        self.tmp_dir = tempfile.TemporaryDirectory()  # pylint: disable=R1732
        self.actual_plots_dir = f"{self.tmp_dir.name}/"
        self.expected_plots_dir = os.path.join(os.path.dirname(__file__), "expected_plots")

    def test_raise_value_error(self):
        """Test if ValueError is raised for wrong argument type."""
        with self.assertRaises(ValueError):
            get_good_linestyles(4)

    def test_get_good_linestyles(self):
        """Test if the default linestyles obtained are the correct ones."""
        with self.subTest("Testing default linestyles"):
            expected_linestyles = [
                "solid",
                (0, (3, 1, 1, 1)),
                (0, (1, 1)),
                (0, (5, 2)),
                (0, (3, 1, 1, 1, 1, 1)),
                (0, (5, 5)),
                (0, (2, 2)),
                "dashdot",
                (0, (5, 10)),
                (0, (1, 10)),
                (0, (3, 10, 1, 10)),
                (0, (3, 10, 1, 10, 1, 10)),
                (0, (3, 5, 1, 5)),
                (0, (3, 5, 1, 5, 1, 5)),
            ]
            actual_linestyles = get_good_linestyles()
            self.assertListEqual(expected_linestyles, actual_linestyles)

        with self.subTest("Test specifying list of names"):
            linestyle_names = ["densely dashed", "dashdotted"]
            expected_linestyles = [(0, (5, 2)), (0, (3, 5, 1, 5))]
            actual_linestyles = get_good_linestyles(linestyle_names)
            self.assertListEqual(expected_linestyles, actual_linestyles)

        with self.subTest("Test case of only one name given (not list)"):
            expected_linestyle = (0, (5, 2))
            actual_linestyle = get_good_linestyles("densely dashed")
            self.assertTupleEqual(expected_linestyle, actual_linestyle)

    def test_linestyles_accepted_by_mpl(self):
        """Test if all the linestyles from get_good_linestyles() are accepted by
        matplotlib.
        """
        test_plot = Line2DPlot()
        for i, linestyle in enumerate(get_good_linestyles()):
            test_plot.add(
                Line2D(
                    np.linspace(0, 10, 10),
                    i * np.linspace(0, 10, 10),
                    linestyle=linestyle,
                )
            )
        test_plot.draw()
        name = "test_linestyles.png"
        test_plot.savefig(f"{self.actual_plots_dir}/{name}")
        # Uncomment line below to update expected image
        # shutil.copy(f"{self.actual_plots_dir}/{name}", f"{self.expected_plots_dir}/{name}")
        self.assertIsNone(
            compare_images(
                f"{self.actual_plots_dir}/{name}",
                f"{self.expected_plots_dir}/{name}",
                tol=1,
            )
        )



"""Variable vs another variable plot."""

from __future__ import annotations

import matplotlib as mpl
import numpy as np
from matplotlib.patches import Rectangle

from puma.plot_base import PlotBase, PlotLineObject
from puma.utils import get_good_colours, get_good_markers, logger
from puma.utils.histogram import hist_ratio


class VarVsVar(PlotLineObject):
    """
    VarVsVar class storing info about curve and allows to calculate ratio w.r.t other
    efficiency plots.
    """

    def __init__(
        self,
        x_var: np.ndarray,
        y_var_mean: np.ndarray,
        y_var_std: np.ndarray,
        x_var_widths: np.ndarray = None,
        key: str | None = None,
        fill: bool = True,
        plot_y_std: bool = True,
        **kwargs,
    ) -> None:
        """Initialise properties of VarVsVar curve object.

        Parameters
        ----------
        x_var : np.ndarray
            Values for x-axis variable, e.g. bin midpoints for binned data
        y_var_mean : np.ndarray
            Mean value for y-axis variable
        y_var_std : np.ndarray
            Std value for y-axis variable
        x_var_widths : np.ndarray, optional
            Widths for x-axis variable, e.g. bin widths for binned data
        key : str, optional
            Identifier for the curve e.g. tagger, by default None
        fill : bool, optional
            Defines do we need to fill box around point, by default True
        plot_y_std : bool, optional
            Defines do we need to plot y_var_std, by default True
        **kwargs : kwargs
            Keyword arguments passed to `PlotLineObject`

        Raises
        ------
        ValueError
            If provided options are not compatible with each other
        """
        super().__init__(**kwargs)
        if len(x_var) != len(y_var_mean):
            raise ValueError(
                f"Length of `x_var` ({len(x_var)}) and `y_var_mean` "
                f"({len(y_var_mean)}) have to be identical."
            )
        if len(x_var) != len(y_var_std):
            raise ValueError(
                f"Length of `x_var` ({len(x_var)}) and `y_var_std` "
                f"({len(y_var_std)}) have to be identical."
            )
        if x_var_widths is not None and len(x_var) != len(x_var_widths):
            raise ValueError(
                f"Length of `x_var` ({len(x_var)}) and `x_var_widths` "
                f"({len(x_var_widths)}) have to be identical."
            )
        self.x_var = np.array(x_var)
        self.x_var_widths = None if x_var_widths is None else np.array(x_var_widths)
        self.y_var_mean = np.array(y_var_mean)
        self.y_var_std = np.array(y_var_std)

        self.key = key
        self.fill = fill
        self.plot_y_std = plot_y_std

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return (
                np.all(self.x_var == other.x_var)
                and np.all(self.y_var_mean == other.y_var_mean)
                and np.all(self.y_var_std == other.y_var_std)
                and self.key == other.key
            )
        return False

    def divide(self, other, inverse: bool = False, method: str = "divide"):
        """Calculate ratio between two class objects.

        Parameters
        ----------
        other : VarVsVar class
            Second VarVsVar object to calculate ratio with
        inverse : bool
            If False the ratio is calculated `this / other`,
            if True the inverse is calculated

        Returns
        -------
        np.ndarray
            Ratio
        np.ndarray
            Ratio error

        Raises
        ------
        ValueError
            If binning is not identical between 2 objects
        """
        if not np.array_equal(self.x_var, other.x_var):
            raise ValueError("The x variables of the two given objects do not match.")
        nom, nom_err = self.y_var_mean, self.y_var_std
        denom, denom_err = other.y_var_mean, other.y_var_std

        ratio, ratio_err = hist_ratio(
            numerator=denom if inverse else nom,
            denominator=nom if inverse else denom,
            numerator_unc=denom_err if inverse else nom_err,
            step=False,
            method=method,
        )
        return (ratio, ratio_err)


class VarVsVarPlot(PlotBase):
    """var_vs_eff plot class."""

    def __init__(self, grid: bool = False, ratio_method: str = "divide", **kwargs) -> None:
        """var_vs_eff plot properties.

        Parameters
        ----------
        grid : bool, optional
            Set the grid for the plots.
        ratio_method:
            Method for ratio calculations. Accepted values: "divide", "root_square_diff".
        **kwargs : kwargs
            Keyword arguments from `puma.PlotObject`

        Raises
        ------
        ValueError
            If incompatible mode given or more than 1 ratio panel requested
        """
        super().__init__(grid=grid, **kwargs)

        self.plot_objects = {}
        self.add_order = []
        self.ratios_objects = {}
        self.reference_object = None
        self.x_var_min = np.inf
        self.x_var_max = -np.inf
        self.inverse_cut = False
        if self.n_ratio_panels > 1:
            raise ValueError("Not more than one ratio panel supported.")
        self.ratio_method = ratio_method
        self.initialise_figure()

    def add(self, curve: VarVsVar, key: str | None = None, reference: bool = False):
        """Adding VarVsVar object to figure.

        Parameters
        ----------
        curve : VarVsVar class
            VarVsVar curve
        key : str, optional
            Unique identifier for VarVsVar curve, by default None
        reference : bool, optional
            If VarVsVar is used as reference for ratio calculation, by default False

        Raises
        ------
        KeyError
            If unique identifier key is used twice
        """
        if key is None:
            key = len(self.plot_objects) + 1
        if key in self.plot_objects:
            raise KeyError(f"Duplicated key {key} already used for unique identifier.")

        self.plot_objects[key] = curve
        self.add_order.append(key)
        # set linestyle
        if curve.linestyle is None:
            curve.linestyle = "-"
        # set colours
        if curve.colour is None:
            curve.colour = get_good_colours()[len(self.plot_objects) - 1]
        # set alpha
        if curve.alpha is None:
            curve.alpha = 0.8
        # set linewidth
        if curve.linewidth is None:
            curve.linewidth = 1.6

        if curve.is_marker is True:
            if curve.marker is None:
                curve.marker = get_good_markers()[len(self.plot_objects)]
            # Set markersize
            if curve.markersize is None:
                curve.markersize = 8
            if curve.markeredgewidth is None:
                curve.markeredgewidth = 2

        # set min and max edges
        if curve.x_var_widths is not None:
            left_edge = curve.x_var - curve.x_var_widths / 2
            right_edge = curve.x_var + curve.x_var_widths / 2
        else:
            left_edge = curve.x_var
            right_edge = curve.x_var
        self.x_var_min = min(self.x_var_min, np.sort(left_edge)[0])
        self.x_var_max = max(self.x_var_max, np.sort(right_edge)[-1])

        if reference:
            logger.debug("Setting roc %s as reference.", key)
            self.set_reference(key)

    def set_reference(self, key: str):
        """Setting the reference roc curves used in the ratios.

        Parameters
        ----------
        key : str
            Unique identifier of roc object
        """
        if self.reference_object is None:
            self.reference_object = key
        else:
            logger.warning(
                "You specified a second curve %s as reference for ratio. "
                "Using it as new reference instead of %s.",
                key,
                self.reference_object,
            )
            self.reference_object = key

    def plot(self, **kwargs):
        """Plotting curves.

        Parameters
        ----------
        **kwargs: kwargs
            Keyword arguments passed to plt.axis.errorbar

        Returns
        -------
        Line2D
            matplotlib Line2D object
        """
        logger.debug("Plotting curves")
        plt_handles = []
        for key in self.add_order:
            elem = self.plot_objects[key]
            error_bar = self.axis_top.errorbar(
                elem.x_var,
                elem.y_var_mean,
                xerr=elem.x_var_widths / 2 if elem.x_var_widths is not None else None,
                yerr=(elem.y_var_std if elem.plot_y_std else np.zeros_like(elem.x_var)),
                color=elem.colour,
                fmt="none",
                label=elem.label,
                alpha=elem.alpha,
                linewidth=elem.linewidth,
                ms=elem.markersize,
                **kwargs,
            )
            # # set linestyle for errorbar
            error_bar[-1][0].set_linestyle(elem.linestyle)
            # Draw markers
            if elem.is_marker is True:
                self.axis_top.scatter(
                    x=elem.x_var,
                    y=elem.y_var_mean,
                    marker=elem.marker,
                    s=elem.markersize**2,
                    color=elem.colour,
                )
            if elem.x_var_widths is not None and elem.fill:
                for x_pos, y_pos, width, height in zip(
                    elem.x_var,
                    elem.y_var_mean,
                    elem.x_var_widths,
                    2 * elem.y_var_std,
                ):
                    self.axis_top.add_patch(
                        Rectangle(
                            xy=(
                                x_pos - width / 2,
                                y_pos - height / 2,
                            ),
                            width=width,
                            height=height,
                            linewidth=0,
                            color=elem.colour,
                            alpha=0.3,
                            zorder=1,
                        )
                    )
            plt_handles.append(
                mpl.lines.Line2D(
                    [],
                    [],
                    color=elem.colour,
                    label=elem.label,
                    linestyle=elem.linestyle,
                    marker=elem.marker,
                    markersize=elem.markersize,
                )
            )
        return plt_handles

    def plot_ratios(self):
        """Plotting ratio curves.

        Raises
        ------
        ValueError
            If no reference curve is defined
        """
        if self.reference_object is None:
            raise ValueError("Please specify a reference curve.")
        for key in self.add_order:
            elem = self.plot_objects[key]
            (ratio, ratio_err) = elem.divide(
                self.plot_objects[self.reference_object], method=self.ratio_method
            )
            error_bar = self.ratio_axes[0].errorbar(
                elem.x_var,
                ratio,
                xerr=elem.x_var_widths / 2 if elem.x_var_widths is not None else None,
                yerr=ratio_err if elem.plot_y_std else np.zeros_like(elem.x_var),
                color=elem.colour,
                fmt="none",
                alpha=elem.alpha,
                linewidth=elem.linewidth,
                ms=elem.markersize,
            )
            # set linestyle for errorbar
            error_bar[-1][0].set_linestyle(elem.linestyle)
            # draw markers
            if elem.is_marker is True:
                self.ratio_axes[0].scatter(
                    x=elem.x_var,
                    y=ratio,
                    marker=elem.marker,
                    color=elem.colour,
                    s=elem.markersize**2,
                )
            if elem.x_var_widths is not None and elem.fill:
                for x_pos, y_pos, width, height in zip(
                    elem.x_var, ratio, elem.x_var_widths, 2 * ratio_err
                ):
                    self.ratio_axes[0].add_patch(
                        Rectangle(
                            xy=(
                                x_pos - width / 2,
                                y_pos - height / 2,
                            ),
                            width=width,
                            height=height,
                            linewidth=0,
                            color=elem.colour,
                            alpha=0.3,
                            zorder=1,
                        )
                    )

    def draw_hline(self, y_val: float):
        """Draw hline in top plot panel.

        Parameters
        ----------
        y_val : float
            y value of the horizontal line
        """
        self.axis_top.hlines(
            y=y_val,
            xmin=self.x_var_min,
            xmax=self.x_var_max,
            colors="black",
            linestyle="dotted",
            alpha=0.5,
        )

    def draw(
        self,
        labelpad: int | None = None,
    ):
        """Draw figure.

        Parameters
        ----------
        labelpad : int, optional
            Spacing in points from the axes bounding box including
            ticks and tick labels, by default "ratio"
        """
        self.set_xlim(
            self.x_var_min if self.xmin is None else self.xmin,
            self.x_var_max if self.xmax is None else self.xmax,
        )
        plt_handles = self.plot()
        if self.n_ratio_panels == 1:
            self.plot_ratios()
        self.set_title()
        self.set_log()
        self.set_y_lim()
        self.set_xlabel()
        self.set_tick_params()
        self.set_ylabel(self.axis_top)

        if self.n_ratio_panels > 0:
            self.set_ylabel(
                self.ratio_axes[0],
                self.ylabel_ratio[0],
                align="center",
                labelpad=labelpad,
            )
        self.make_legend(plt_handles, ax_mpl=self.axis_top)
        self.plotting_done = True
        if self.apply_atlas_style is True:
            self.atlasify()



from __future__ import annotations

import numpy as np

from puma.utils.confusion_matrix import confusion_matrix


def precision_recall_scores_per_class(
    targets: np.ndarray,
    predictions: np.ndarray,
    sample_weights: np.ndarray | None = None,
) -> np.ndarray:
    """
    Compute the per-class precision and recall scores of a classification,
    given the target and predicted labels. Return two arrays where the entry ``i`` is the
    precision or recall score of the class ``i`` respectively.
    The precision score is defined, for each class, as ``tp / (tp + fp)``,
    where ``tp`` is the number of true positives and ``fp`` is the number of false positives.
    The recall score is defined, for each class, as ``tp / (tp + fn)`` where ``tp`` is the number of
    true positives and ``fn`` is the number of false negatives.

    Parameters
    ----------
    targets : 1d np.ndarray
        target labels
    predictions : 1d np.ndarray
        predicted labels (output of the classifier)
    sample_weights : np.ndarray, optional
        Weight of each sample; if None, each sample weights the same. Defaults to None.

    Returns
    -------
    np.ndarray : the per-class precision score.
    np.ndarray : the per-class recall score.

    Example
    --------
    >>> targets = np.array([2, 0, 2, 2, 0, 1])
    >>> predictions = np.array([0, 0, 2, 2, 0, 2])
    >>> weights = np.array([1, 0.5, 0.5, 1, 0.2, 1])
    >>> p,r = precision_recall_score_per_class(targets, predictions, sample_weights=weights)
    >>> p
    [0.41176471, 1.0, 1.0]
    >>> r
    [1.0, 0.0, 0.6]
    """
    cm = confusion_matrix(targets, predictions, sample_weights=sample_weights, normalize=None)
    tp = np.diag(cm)
    with np.errstate(all="warn"):
        tpfp = np.sum(cm, axis=0)
        tpfn = np.sum(cm, axis=1)
        precision = tp / tpfp
        recall = tp / tpfn

    return np.nan_to_num(precision), np.nan_to_num(recall)



"""Unit test script for the functions in line_plot_2d.py."""

from __future__ import annotations

import os
import shutil  # noqa: F401
import tempfile
import unittest

import numpy as np
from matplotlib.testing.compare import compare_images

from puma import Line2D, Line2DPlot
from puma.utils import logger, set_log_level

set_log_level(logger, "DEBUG")


class Line2DTestCase(unittest.TestCase):
    """Test class for the puma.line_plot_2d functions."""

    def test_wrong_inputs_xvalues(self):
        """Test if providing wrong input type to Line2D raises ValueError."""
        with self.assertRaises(TypeError):
            Line2D(
                x_values="Test",
                y_values=5,
            )

    def test_differnt_input_types(self):
        """Test if providing different input types raises ValueError."""
        with self.assertRaises(ValueError):
            Line2D(
                x_values=[1, 2, 3],
                y_values=np.array([1, 2, 3]),
            )

    def test_empty_input(self):
        """Test if ValueError is raised when one of the input values is zero."""
        with self.assertRaises(ValueError):
            Line2D(
                x_values=[],
                y_values=[1, 2],
            )

        with self.assertRaises(ValueError):
            Line2D(
                x_values=[1, 2],
                y_values=[],
            )

    def test_different_input_shapes(self):
        """Test if ValueError is raised when different lengths given."""
        with self.assertRaises(ValueError):
            Line2D(
                x_values=[1, 2, 3],
                y_values=[1, 2],
            )


class Line2DPlotTestCase(unittest.TestCase):
    """Test class for puma.Line2DPlot."""

    def setUp(self):
        """Set up values needed."""
        # Line values
        self.x_values = np.arange(0.001, 1, 0.001)
        self.y_values = np.arange(0.001, 1, 0.001)

        # Marker values
        self.marker_x = 0.5
        self.marker_y = 0.5

        # Set up directories for comparison plots
        self.tmp_dir = tempfile.TemporaryDirectory()  # pylint: disable=R1732
        self.actual_plots_dir = f"{self.tmp_dir.name}"
        self.expected_plots_dir = os.path.join(os.path.dirname(__file__), "expected_plots")

    def test_basic(self):
        """Test the basic functions of Line2DPlot."""
        frac_plot = Line2DPlot(n_ratio_panels=0)

        # Add line
        frac_plot.add(
            Line2D(
                x_values=self.x_values,
                y_values=self.y_values,
                label="Tagger 1",
                colour="r",
                linestyle="-",
            )
        )

        # Add marker
        frac_plot.add(
            Line2D(
                x_values=self.marker_x,
                y_values=self.marker_y,
                colour="r",
                marker="x",
                label="Marker label",
                markersize=15,
                markeredgewidth=2,
            ),
            is_marker=True,
        )

        frac_plot.xlabel = "Test_x_label"
        frac_plot.ylabel = "Test_y_label"

        # Define a plot name
        name = "test_line2d_all_params_given.png"

        # Draw and save the plot
        frac_plot.draw()
        frac_plot.savefig(f"{self.actual_plots_dir}/{name}")

        # Uncomment line below to update expected image
        # shutil.copy(f"{self.actual_plots_dir}/{name}", f"{self.expected_plots_dir}/{name}")

        # Check
        self.assertIsNone(
            compare_images(
                f"{self.actual_plots_dir}/{name}",
                f"{self.expected_plots_dir}/{name}",
                tol=1,
            )
        )

    def test_grid_off(self):
        """Test the basic functions of Line2DPlot."""
        frac_plot = Line2DPlot(n_ratio_panels=0, grid=False)

        # Add line
        frac_plot.add(
            Line2D(
                x_values=self.x_values,
                y_values=self.y_values,
                label="Tagger 1",
                colour="r",
                linestyle="-",
            )
        )

        # Add marker
        frac_plot.add(
            Line2D(
                x_values=self.marker_x,
                y_values=self.marker_y,
                colour="r",
                marker="x",
                label="Marker label",
                markersize=15,
                markeredgewidth=2,
            ),
            is_marker=True,
        )

        frac_plot.xlabel = "Test_x_label"
        frac_plot.ylabel = "Test_y_label"

        # Define a plot name
        name = "test_line2d_grid_off.png"

        # Draw and save the plot
        frac_plot.draw()
        frac_plot.savefig(f"{self.actual_plots_dir}/{name}")

        # Uncomment line below to update expected image
        # shutil.copy(f"{self.actual_plots_dir}/{name}", f"{self.expected_plots_dir}/{name}")

        # Check
        self.assertIsNone(
            compare_images(
                f"{self.actual_plots_dir}/{name}",
                f"{self.expected_plots_dir}/{name}",
                tol=1,
            )
        )

    def test_no_param_set(self):
        """Test the basic functions of Line2DPlot."""
        frac_plot = Line2DPlot(n_ratio_panels=0)

        # Add line
        frac_plot.add(
            Line2D(
                x_values=self.x_values,
                y_values=self.y_values,
                label="Tagger 1",
                colour=None,
                linestyle=None,
            )
        )

        # Add marker
        frac_plot.add(
            Line2D(
                x_values=self.marker_x,
                y_values=self.marker_y,
                colour=None,
                marker=None,
                label=None,
                markersize=None,
                markeredgewidth=None,
            ),
            is_marker=True,
        )

        frac_plot.xlabel = "Test_x_label"
        frac_plot.ylabel = "Test_y_label"
        frac_plot.logx = True
        frac_plot.logy = True

        # Define a plot name
        name = "test_line2d_no_params_given.png"

        # Draw and save the plot
        frac_plot.draw()
        frac_plot.savefig(f"{self.actual_plots_dir}/{name}")

        # Uncomment line below to update expected image
        # shutil.copy(f"{self.actual_plots_dir}/{name}", f"{self.expected_plots_dir}/{name}")

        # Check
        self.assertIsNone(
            compare_images(
                f"{self.actual_plots_dir}/{name}",
                f"{self.expected_plots_dir}/{name}",
                tol=1,
            )
        )

    def test_double_key(self):
        """Test the basic functions of Line2DPlot."""
        frac_plot = Line2DPlot(n_ratio_panels=0)

        # Add line
        frac_plot.add(
            Line2D(
                x_values=self.x_values,
                y_values=self.y_values,
                label="Tagger 1",
                colour="r",
                linestyle="-",
            ),
            key=1,
        )

        with self.assertRaises(KeyError):
            # Add marker
            frac_plot.add(
                Line2D(
                    x_values=self.marker_x,
                    y_values=self.marker_y,
                    colour="r",
                    marker="x",
                    label="Marker label",
                    markersize=15,
                    markeredgewidth=2,
                ),
                is_marker=True,
                key=1,
            )



"""Classes for 2D line plots."""

from __future__ import annotations

import matplotlib as mpl
import numpy as np
import pandas as pd

from puma.plot_base import PlotBase, PlotLineObject
from puma.utils import get_good_colours, get_good_markers, logger


class Line2D(PlotLineObject):
    """Line2D class storing info about the x and y values and style."""

    def __init__(
        self,
        x_values: np.ndarray,
        y_values: np.ndarray,
        **kwargs,
    ) -> None:
        """Initialise properties of Line2D object.

        Parameters
        ----------
        x_values : np.ndarray
            x values of the curve
        y_values : np.ndarray
            y values of the curve
        **kwargs : kwargs
            kwargs passed to `PlotLineObject`

        Raises
        ------
        ValueError
            If the dtype of x_values and y_values is different.
        ValueError
            If provided x_values array is empty.
        ValueError
            If provided y_values array is empty.
        ValueError
            If provided x_values and y_values arrays have different
            shapes.
        ValueError
            If an invalid type was given for x_values
        """
        super().__init__(**kwargs)

        # Check input dtype
        if isinstance(x_values, (np.ndarray, list, int, float, pd.Series)):
            if type(x_values) != type(y_values):  # pylint: disable=C0123
                raise ValueError(
                    "Invalid types of input given! Both must be one of the following: "
                    "numpy.ndarray, list, int, float, pandas.Series \n"
                    f"You specified: x_values: {type(x_values)} , "
                    f"y_values: {type(y_values)}"
                )

            if isinstance(x_values, (int, float)) and isinstance(y_values, (int, float)):
                # Convert input into numpy array
                x_values = np.array([x_values])
                y_values = np.array([y_values])

            # Convert input into numpy array
            x_values = np.array(x_values)
            y_values = np.array(y_values)

            # Check that given arrays/lists are not empty
            if len(x_values) == 0:
                raise ValueError("Provided x_values is empty!")

            if len(y_values) == 0:
                raise ValueError("Provided y_values is empty!")

            # Check that both inputs have the same dimension
            if len(x_values) != len(y_values):
                raise ValueError(
                    "x_values and y_values have different dimensionalities! "
                    f"x_values: {len(x_values)}, y_values: {len(y_values)}"
                )

        else:
            raise TypeError(
                "Invalid type of input data. Allowed values are numpy.ndarray, list," " int, float"
            )

        # Set inputs as attributes
        self.x_values = x_values
        self.y_values = y_values

        # Set key to None. Will be defined when plotting starts
        self.key = None


class Line2DPlot(PlotBase):
    """Line2DPlot plot class for basic x-y line plots."""

    def __init__(
        self,
        logy: bool = False,
        grid: bool = True,
        **kwargs,
    ) -> None:
        """Plot properties.

        Parameters
        ----------
        logy : bool, optional
            Decide, if the y-axis of the plot will be in log, by default False
        grid : bool, optional
            Set the grid for the plots.
        **kwargs : kwargs
            Keyword arguments from `puma.PlotObject`
        """
        super().__init__(grid=grid, **kwargs)

        # Set inputs as attributes
        self.logy = logy

        # Init needed lists and dicts
        self.plot_objects = {}
        self.add_order = []

        self.initialise_figure()

    def add(
        self,
        curve: object,
        key: str | None = None,
        is_marker: bool = False,
    ):
        """Adding `puma.Line2D` object to figure.

        Parameters
        ----------
        curve : puma.line_plot_2D.Line2D
            Line2D object
        key : str, optional
            Unique identifier for the curve, by default None
        is_marker : bool, optional
            Defines if this is a marker (True) or a line (False). By default False.

        Raises
        ------
        KeyError
            If unique identifier key is used twice
        """
        # If key not defined, set it to a numerical value
        if key is None:
            key = len(self.plot_objects) + 1

        # Check that key is not double used
        if key in self.plot_objects:
            raise KeyError(f"Duplicated key {key} already used for unique identifier.")

        # Add key to Line2D object
        curve.key = key
        logger.debug("Adding Line2D object with key %s", key)

        # Set alpha
        if curve.alpha is None:
            curve.alpha = 1

        if is_marker is False:
            # Set colours
            if curve.colour is None:
                curve.colour = get_good_colours()[len(self.plot_objects)]
            curve.is_marker = False
            # Set linestyle
            if curve.linestyle is None:
                curve.linestyle = "-"
            # Set linewidth
            if curve.linewidth is None:
                curve.linewidth = 1.6

        else:
            curve.is_marker = True
            # Set colour of the marker the same as the last line plotted
            if curve.colour is None:
                curve.colour = self.plot_objects[len(self.plot_objects)].colour
            # Set markerstyle
            if curve.marker is None:
                curve.marker = get_good_markers()[len(self.plot_objects)]
            # Set markersize
            if curve.markersize is None:
                curve.markersize = 15
            # Set markersize
            if curve.markeredgewidth is None:
                curve.markeredgewidth = 2
            # Set the linestyle for markers to None as string
            # to suppress the line in the legend so only the marker
            # is shown in the legend
            curve.linestyle = "None"

        self.plot_objects[key] = curve
        self.add_order.append(key)

    def plot(self, **kwargs):
        """Plotting curves. Plot objects are drawn in the same order as they
        were added to the plot.

        Parameters
        ----------
        **kwargs: kwargs
            Keyword arguments passed to matplotlib.axes.Axes.plot()

        Returns
        -------
        Line2D
            matplotlib Line2D object
        """
        plt_handles = []

        # Loop over all plot objects and plot them
        for key in self.add_order:
            elem = self.plot_objects[key]

            self.axis_top.plot(
                elem.x_values,
                elem.y_values,
                color=elem.colour,
                label=elem.label,
                alpha=elem.alpha,
                linewidth=elem.linewidth,
                linestyle=elem.linestyle,
                marker=elem.marker,
                markersize=elem.markersize,
                markeredgewidth=elem.markeredgewidth,
                **kwargs,
            )

            plt_handles.append(
                mpl.lines.Line2D(
                    [],
                    [],
                    color=elem.colour,
                    label=elem.label,
                    linestyle=elem.linestyle,
                    marker=elem.marker,
                )
            )

        self.plotting_done = True
        return plt_handles

    def draw(self):
        """Draw figure."""
        plt_handles = self.plot()

        # Make the legend
        self.make_legend(plt_handles, ax_mpl=self.axis_top)

        self.set_title()
        self.set_log()
        self.set_y_lim()
        self.set_xlabel()
        self.set_tick_params()
        self.set_ylabel(self.axis_top)

        # Apply atlas style if defined
        if self.apply_atlas_style:
            self.atlasify()



"""Unit test script for the functions in hlplots/tagger.py."""

from __future__ import annotations

import tempfile
import unittest
from pathlib import Path

import h5py
import numpy as np
import pandas as pd
from ftag import Cuts
from numpy.lib.recfunctions import structured_to_unstructured as s2u
from numpy.lib.recfunctions import unstructured_to_structured as u2s

from puma.hlplots import Tagger
from puma.utils import logger, set_log_level

set_log_level(logger, "DEBUG")


class TaggerBasisTestCase(unittest.TestCase):
    """Test class for the Tagger class."""

    def test_empty_string_tagger_name(self):
        """Test empty string as model name."""
        tagger = Tagger("")
        self.assertEqual(tagger.name, "")

    def test_more_tagger_args(self):
        tagger = Tagger("dummy", sample_path="dummy_path", cuts=[("pt", ">", 20)])
        assert isinstance(tagger.sample_path, Path)
        assert isinstance(tagger.cuts, Cuts)
        assert repr(tagger) == "dummy (dummy)"

    def test_wrong_template(self):
        """Test wrong template."""
        template_wrong = {"test": 1}
        with self.assertRaises(TypeError):
            Tagger("dummy", **template_wrong)  # pylint: disable=W0212,E1123

    def test_label_template(self):
        """Test template with label."""
        template_label = {"label": 1.5}
        tagger = Tagger("dummy", **template_label)  # pylint: disable=W0212
        self.assertEqual(tagger.label, 1.5)

    def test_n_jets(self):
        """Test if number of n_jets correctly calculated."""
        tagger = Tagger("dummy", output_flavours=["ujets", "cjets", "bjets"])
        labels = np.concatenate([np.zeros(80), np.ones(5) * 4, np.ones(15) * 5])
        tagger.labels = np.array(labels, dtype=[("HadronConeExclTruthLabelID", "i4")])
        with self.subTest():
            self.assertEqual(tagger.n_jets("ujets"), 80)
        with self.subTest():
            self.assertEqual(tagger.n_jets("cjets"), 5)
        with self.subTest():
            self.assertEqual(tagger.n_jets("bjets"), 15)
        assert np.sum(tagger.is_flav("ujets")) == 3160


class TaggerScoreExtractionTestCase(unittest.TestCase):
    """Test extract_tagger_scores function in Tagger class."""

    def setUp(self) -> None:
        """Set up for tests."""
        self.df_dummy = pd.DataFrame({
            "dummy_pc": np.zeros(10),
            "dummy_pu": np.ones(10),
            "dummy_pb": np.zeros(10),
        })
        self.scores_expected = np.column_stack((np.ones(10), np.zeros(10), np.zeros(10)))

    def test_wrong_source_type(self):
        """Test using wrong source type."""
        tagger = Tagger("dummy")
        with self.assertRaises(ValueError):
            tagger.extract_tagger_scores(self.df_dummy, source_type="dummy")

    def test_data_frame_path_no_key(self):
        """Test passing data frame path but no key."""
        tagger = Tagger("dummy")
        with self.assertRaises(ValueError):
            tagger.extract_tagger_scores(self.df_dummy, source_type="data_frame_path")

    def test_data_frame(self):
        """Test passing data frame."""
        tagger = Tagger("dummy")
        tagger.extract_tagger_scores(self.df_dummy)
        np.testing.assert_array_equal(s2u(tagger.scores), self.scores_expected)

    def test_data_frame_path(self):
        """Test passing data frame path."""
        tagger = Tagger("dummy")
        with tempfile.TemporaryDirectory() as tmp_dir:
            file_name = f"{tmp_dir}/dummy_df.h5"
            self.df_dummy.to_hdf(file_name, key="dummy_tagger")

            tagger.extract_tagger_scores(
                file_name, key="dummy_tagger", source_type="data_frame_path"
            )
        np.testing.assert_array_equal(s2u(tagger.scores), self.scores_expected)

    def test_h5_structured_numpy_path(self):
        """Test passing structured h5 path."""
        tagger = Tagger("dummy")
        with tempfile.TemporaryDirectory() as tmp_dir:
            file_name = f"{tmp_dir}/dummy_df.h5"
            with h5py.File(file_name, "w") as f_h5:
                f_h5.create_dataset(data=self.df_dummy.to_records(), name="dummy_tagger")

            tagger.extract_tagger_scores(file_name, key="dummy_tagger", source_type="h5_file")
        np.testing.assert_array_equal(s2u(tagger.scores), self.scores_expected)

    def test_structured_array(self):
        """Test passing structured numpy array."""
        tagger = Tagger("dummy")
        tagger.extract_tagger_scores(
            self.df_dummy.to_records(),
            key="dummy_tagger",
            source_type="structured_array",
        )
        np.testing.assert_array_equal(s2u(tagger.scores), self.scores_expected)


class TaggerTestCase(unittest.TestCase):
    """Test class for the Tagger class."""

    def setUp(self) -> None:
        """Set up for tests."""
        scores = np.column_stack((np.ones(10), np.ones(10), np.ones(10)))
        self.scores = u2s(
            scores,
            dtype=[("dummy_pu", "f4"), ("dummy_pc", "f4"), ("dummy_pb", "f4")],
        )

    def test_errors(self):
        tagger = Tagger("dummy")
        tagger.scores = self.scores
        with self.assertRaises(ValueError):
            tagger.discriminant("qcd")
        with self.assertRaises(ValueError):
            tagger.discriminant("ujets")

    def test_disc_b_calc(self):
        """Test b-disc calculation."""
        tagger = Tagger("dummy", fxs={"fc": 0.5})
        tagger.scores = self.scores
        discs = tagger.discriminant("bjets")
        np.testing.assert_array_equal(discs, np.zeros(10))

    def test_disc_c_calc(self):
        """Test c-disc calculation."""
        tagger = Tagger("dummy", fxs={"fb": 0.5})
        tagger.scores = self.scores
        discs = tagger.discriminant("cjets")
        np.testing.assert_array_equal(discs, np.zeros(10))

    def test_disc_hbb_calc(self):
        """Test hbb-disc calculation."""
        from ftag import Flavours as F

        tagger = Tagger(
            "dummy",
            fxs={"fhcc": 0.1, "ftop": 0.1},
            output_flavours=[F["hbb"], F["hcc"], F["top"], F["qcd"]],
        )
        tagger.scores = u2s(
            np.column_stack((np.ones(10) * 2, np.ones(10), np.ones(10), np.ones(10))),
            dtype=[
                ("dummy_phbb", "f4"),
                ("dummy_phcc", "f4"),
                ("dummy_ptop", "f4"),
                ("dummy_pqcd", "f4"),
            ],
        )
        discs = tagger.discriminant("hbb")
        np.testing.assert_array_equal(discs, np.ones([10]) * 0.6931471824645996)


class TaggerAuxTaskTestCase(unittest.TestCase):
    """Test class for aux task functionality in Tagger class."""

    def test_aux_variables_vertexing(self):
        """Test vertexing aux task variable retrieval. Includes special cases for SV1
        and JF taggers.
        """
        default_tagger = Tagger("dummy", aux_tasks=["vertexing"])
        SV1_tagger = Tagger("SV1", aux_tasks=["vertexing"])
        JF_tagger = Tagger("JF", aux_tasks=["vertexing"])
        self.assertEqual(default_tagger.aux_variables["vertexing"], "dummy_aux_VertexIndex")
        self.assertEqual(SV1_tagger.aux_variables["vertexing"], "SV1VertexIndex")
        self.assertEqual(JF_tagger.aux_variables["vertexing"], "JFVertexIndex")

    def test_aux_variables_track_origin(self):
        """Test track_origin aux task variable retrieval."""
        tagger = Tagger("dummy", aux_tasks=["track_origin"])
        self.assertEqual(tagger.aux_variables["track_origin"], "dummy_aux_TrackOrigin")

    def test_aux_variables_undefined(self):
        """Test undefined aux task variable retrieval."""
        tagger = Tagger("dummy", aux_tasks=["dummy"])
        with self.assertRaises(ValueError):
            tagger.aux_variables["dummy"]


class TaggerProbsTestCase(unittest.TestCase):
    """Test class for the probs method in Tagger class."""

    def setUp(self) -> None:
        """Set up for tests."""
        scores = np.column_stack((np.ones(10), np.zeros(10), np.ones(10) * 5))
        self.scores = u2s(
            scores,
            dtype=[("dummy_pu", "f4"), ("dummy_pc", "f4"), ("dummy_pb", "f4")],
        )
        self.tagger = Tagger("dummy")
        self.tagger.scores = self.scores
        self.tagger.labels = np.array(
            np.zeros(len(self.scores)), dtype=[("HadronConeExclTruthLabelID", "i4")]
        )

    def test_probs_without_label_flavour(self):
        """Test probs method without label_flavour parameter."""
        prob_flavour = "bjets"
        expected_probs = np.ones(10) * 5
        np.testing.assert_array_equal(self.tagger.probs(prob_flavour), expected_probs)

    def test_probs_with_label_flavour(self):
        """Test probs method with label_flavour parameter."""
        prob_flavour = "bjets"
        label_flavour = "cjets"
        expected_probs = np.zeros(0)
        np.testing.assert_array_equal(
            self.tagger.probs(prob_flavour, label_flavour), expected_probs
        )
        prob_flavour = "cjets"
        label_flavour = "ujets"
        expected_probs = np.zeros(10)
        np.testing.assert_array_equal(
            self.tagger.probs(prob_flavour, label_flavour), expected_probs
        )



"""Helper function for histogram handling."""

from __future__ import annotations

import numpy as np

from puma.utils.logging import logger


def save_divide(
    numerator,
    denominator,
    default: float = 1.0,
):
    """
    Division using numpy divide function returning default value in cases where
    denominator is 0.

    Parameters
    ----------
    numerator: array_like, int, float
        Numerator in the ratio calculation.
    denominator: array_like, int, float
        Denominator in the ratio calculation.
    default: float
        Default value which is returned if denominator is 0.

    Returns
    -------
    ratio: array_like, float
    """
    if isinstance(numerator, (int, float, np.number)) and isinstance(
        denominator, (int, float, np.number)
    ):
        output_shape = 1
    else:
        try:
            output_shape = denominator.shape
        except AttributeError:
            output_shape = numerator.shape

    ratio = np.divide(
        numerator,
        denominator,
        out=np.ones(
            output_shape,
            dtype=float,
        )
        * default,
        where=(denominator != 0),
    )
    if output_shape == 1:
        return float(ratio)
    return ratio


def hist_w_unc(
    arr,
    bins,
    filled: bool = False,
    bins_range=None,
    normed: bool = True,
    weights: np.ndarray = None,
    bin_edges: np.ndarray = None,
    sum_squared_weights: np.ndarray = None,
    underoverflow: bool = False,
):
    """
    Computes histogram and the associated statistical uncertainty.

    Parameters
    ----------
    arr : array_like
        Input data. The histogram is computed over the flattened array.
    bins : int or sequence of scalars or str
        `bins` parameter from `np.histogram`
    bins_range : tuple, optional
        `range` parameter from `np.histogram`. This is ignored if `bins` is array like,
        because then the entries of `bins` are used as bin edges.
    normed : bool, optional
        If True (default) the calculated histogram is normalised to an integral
        of 1.
    weights : np.ndarray, optional
        Weights for the input data. Has to be an array of same length as the input
        data with a weight for each entry. If not specified, weight 1 will be given
        to each entry. The uncertainty of bins with weighted entries is
        sqrt(sum_i{w_i^2}) where w_i are the weights of the entries in this bin.
        By default None.
    underoverflow : bool, optional
        Option to include under- and overflow values in outermost bins.

    Returns
    -------
    bin_edges : array of dtype float
        Return the bin edges (length(hist)+1)
    hist : numpy array
        The values of the histogram. If normed is true (default), returns the
        normed counts per bin
    unc : numpy array
        Statistical uncertainty per bin.
        If normed is true (default), returns the normed values.
    band : numpy array
        lower uncertainty band location: hist - unc
        If normed is true (default), returns the normed values.
    """
    if weights is None:
        weights = np.ones(len(arr))

    # Check if there are nan values in the input values
    nan_mask = np.isnan(arr)
    if np.sum(nan_mask) > 0:
        logger.warning("Histogram values contain %i nan values!", np.sum(nan_mask))
        # Remove nan values
        arr = arr[~nan_mask]
        weights = weights[~nan_mask]

    # Check if there are inf values in the input values
    inf_mask = np.isinf(arr)
    if np.sum(inf_mask) > 0:
        logger.warning("Histogram values contain %i +-inf values!", np.sum(inf_mask))

    # If the histogram is not already filled we need to produce the histogram counts
    # and bin edges
    if not filled:
        # Calculate the counts and the bin edges
        counts, bin_edges = np.histogram(arr, bins=bins, range=bins_range, weights=weights)

        # calculate the uncertainty with sum of squared weights (per bin, so we use
        # np.histogram again here)
        unc = np.sqrt(np.histogram(arr, bins=bins, range=bins_range, weights=weights**2)[0])

        if underoverflow:
            # add two dummy bins (from outermost bins to +-infinity)
            bins_with_overunderflow = np.hstack([
                np.array([-np.inf]),
                bin_edges,
                np.array([np.inf]),
            ])
            # recalculate the histogram with this adjusted binning
            counts, _ = np.histogram(arr, bins=bins_with_overunderflow, weights=weights)
            counts[1] += counts[0]  # add underflow values to underflow bin
            counts[-2] += counts[-1]  # add overflow values to overflow bin
            counts = counts[1:-1]  # remove dummy bins

            # calculate the sum of squared weights
            sum_squared_weights = np.histogram(
                arr, bins=bins_with_overunderflow, weights=weights**2
            )[0]

            # add sum of squared weights from under/overflow values
            # to under/overflow bin
            sum_squared_weights[1] += sum_squared_weights[0]
            sum_squared_weights[-2] += sum_squared_weights[-1]
            # remove dummy bins
            sum_squared_weights = sum_squared_weights[1:-1]

            # uncertainty is sqrt(sum_squared_weights)
            unc = np.sqrt(sum_squared_weights)

        if normed:
            sum_of_weights = float(np.sum(weights))
            counts = save_divide(counts, sum_of_weights, 0)
            unc = save_divide(unc, sum_of_weights, 0)

    # If the histogram is already filled then the uncertainty is computed
    # differently
    else:
        if sum_squared_weights is not None:
            sum_squared_weights = np.array(sum_squared_weights)[~nan_mask]
            unc = np.sqrt(sum_squared_weights)
        else:
            unc = np.sqrt(arr)  # treat arr as bin heights (counts)

        counts = arr

        if normed:
            counts_sum = float(np.sum(counts))
            counts = save_divide(counts, counts_sum, 0)
            unc = save_divide(unc, counts_sum, 0)

    # regardless of if the histogram is filled
    band = counts - unc
    hist = counts

    return bin_edges, hist, unc, band


def hist_ratio(
    numerator,
    denominator,
    numerator_unc,
    step: bool = True,
    method: str = "divide",
):
    """
    Calculate the ratio of the given bincounts and
    returns the input for a step function that plots the ratio.

    Parameters
    ----------
    numerator : array_like
        Numerator in the ratio calculation.
    denominator : array_like
        Denominator in the ratio calculation.
    numerator_unc : array_like
        Uncertainty of the numerator.
    step : bool
        if True duplicates first bin to match with step plotting function,
        by default True
    method : str
        Selects the method by which the ratio should be calculated.
        "divide" calculates the ratio as the division of the numerator by the denominator.
        "root_square_diff" calculates the Root Square Difference
        between the numerator and the denominator.
        by default "divide"


    Returns
    -------
    step_ratio : array_like
        Ratio returning 1 in case the denominator is 0.
    step_ratio_unc : array_like
        Stat. uncertainty of the step_ratio

    Raises
    ------
    AssertionError
        If inputs don't have the same shape.

    """
    numerator, denominator, numerator_unc = (
        np.array(numerator),
        np.array(denominator),
        np.array(numerator_unc),
    )
    if numerator.shape != denominator.shape:
        raise AssertionError("Numerator and denominator don't have the same legth")
    if numerator.shape != numerator_unc.shape:
        raise AssertionError("Numerator and numerator_unc don't have the same legth")

    if method == "divide":
        step_ratio = save_divide(numerator, denominator, 1 if step else np.inf)
        # Calculate ratio uncertainty
        step_unc = save_divide(numerator_unc, denominator, default=0 if step else np.inf)
    elif method == "root_square_diff":
        step_ratio = np.multiply(
            np.sign(numerator - denominator), np.sqrt(np.abs(numerator**2 - denominator**2))
        )
        # Calculate ratio uncertainty
        step_unc = np.zeros_like(step_ratio)
        step_unc = np.divide(
            np.multiply(numerator, numerator_unc),
            np.sqrt(np.abs(numerator**2 - denominator**2)),
            where=(numerator - denominator != 0),
        )
    else:
        raise ValueError("'method' can only be 'divide' or 'root_square_diff'.")

    if step:
        # Add an extra bin in the beginning to have the same binning as the input
        # Otherwise, the ratio will not be exactly above each other (due to step)
        step_ratio = np.append(np.array([step_ratio[0]]), step_ratio)
        step_unc = np.append(np.array([step_unc[0]]), step_unc)

    return step_ratio, step_unc



from __future__ import annotations

import unittest

import numpy as np

from puma.utils import logger, set_log_level
from puma.utils.precision_recall_scores import precision_recall_scores_per_class

set_log_level(logger, "DEBUG")


class ConfusionMatrixTestCase(unittest.TestCase):
    def test_precision_unweighted(self):
        targets = np.array([2, 0, 2, 2, 0, 1])
        predictions = np.array([0, 0, 2, 2, 0, 2])
        precision, recall = precision_recall_scores_per_class(targets, predictions)
        expected_precision = np.array([0.66666667, 0.0, 0.66666667])
        expected_recall = np.array([1.0, 0.0, 0.66666667])
        np.testing.assert_array_almost_equal(expected_precision, precision)
        np.testing.assert_array_almost_equal(expected_recall, recall)

    def test_precision_weighted(self):
        targets = np.array([2, 0, 2, 2, 0, 1])
        predictions = np.array([0, 0, 2, 2, 0, 2])
        weights = np.array([1, 0.5, 0.5, 1, 0.2, 1])
        precision, recall = precision_recall_scores_per_class(targets, predictions, weights)
        expected_precision = np.array([0.41176471, 0.0, 0.6])
        expected_recall = np.array([1.0, 0.0, 0.6])
        np.testing.assert_array_almost_equal(expected_precision, precision)
        np.testing.assert_array_almost_equal(expected_recall, recall)



"""Unit test script for the functions in var_vs_var.py."""

from __future__ import annotations

import os
import shutil  # noqa: F401
import tempfile
import unittest

import numpy as np
from matplotlib.testing.compare import compare_images

from puma import VarVsVar, VarVsVarPlot
from puma.utils.logging import logger, set_log_level

set_log_level(logger, "DEBUG")


class VarVsVarTestCase(unittest.TestCase):
    """Test class for the puma.var_vs_var functions."""

    def setUp(self):
        self.x_var = np.linspace(100, 250, 20)
        self.y_var_mean = np.exp(-np.linspace(6, 10, 20)) * 10e3
        self.y_var_std = np.sin(self.y_var_mean)

    def test_var_vs_var_init_wrong_mean_shape(self):
        """Test var_vs_var init."""
        with self.assertRaises(ValueError):
            VarVsVar(np.ones(4), np.ones(5), np.ones(5))

    def test_var_vs_var_init_wrong_y_mean_std_shape(self):
        """Test var_vs_var init."""
        with self.assertRaises(ValueError):
            VarVsVar(np.ones(4), np.ones(4), np.ones(5))

    def test_var_vs_var_init_wrong_x_mean_widths_shape(self):
        """Test var_vs_var init."""
        with self.assertRaises(ValueError):
            VarVsVar(np.ones(4), np.ones(4), np.ones(4), x_var_widths=np.ones(5))

    def test_var_vs_var_init(self):
        """Test var_vs_var init."""
        VarVsVar(
            np.ones(6),
            np.ones(6),
            np.ones(6),
            x_var_widths=np.ones(6),
            key="test",
            fill=True,
            plot_y_std=False,
        )

    def test_var_vs_var_eq(self):
        """Test var_vs_var eq."""
        var_plot = VarVsVar(
            np.ones(6),
            np.ones(6),
            np.ones(6),
            x_var_widths=np.ones(6),
            key="test",
            fill=True,
            plot_y_std=False,
        )
        self.assertEqual(var_plot, var_plot)

    def test_var_vs_var_eq_different_classes(self):
        """Test var_vs_var eq."""
        var_plot = VarVsVar(
            np.ones(6),
            np.ones(6),
            np.ones(6),
            x_var_widths=np.ones(6),
            key="test",
            fill=True,
            plot_y_std=False,
        )
        self.assertNotEqual(var_plot, np.ones(6))

    def test_var_vs_var_divide_same(self):
        """Test var_vs_var divide."""
        var_plot = VarVsVar(
            x_var=self.x_var,
            y_var_mean=self.y_var_mean,
            y_var_std=self.y_var_std,
        )
        np.testing.assert_array_almost_equal(var_plot.divide(var_plot)[0], np.ones(20))

    def test_var_vs_var_divide_different_shapes(self):
        """Test var_vs_eff divide."""
        var_plot = VarVsVar(
            x_var=self.x_var,
            y_var_mean=self.y_var_mean,
            y_var_std=self.y_var_std,
        )
        var_plot_comp = VarVsVar(
            x_var=np.repeat(self.x_var, 2),
            y_var_mean=np.repeat(self.y_var_mean, 2),
            y_var_std=np.repeat(self.y_var_std, 2),
        )
        with self.assertRaises(ValueError):
            var_plot.divide(var_plot_comp)


class VarVsVarPlotTestCase(unittest.TestCase):
    """Test class for the puma.var_vs_var_plot."""

    def setUp(self):
        # Set up temp directory for comparison plots
        self.tmp_dir = tempfile.TemporaryDirectory()  # pylint:disable=R1732
        self.actual_plots_dir = f"{self.tmp_dir.name}/"
        self.expected_plots_dir = os.path.join(os.path.dirname(__file__), "expected_plots")
        np.random.seed(42)
        n_random = 21

        # background (same for both taggers)
        self.x_var = np.linspace(0, 250, num=n_random)
        self.y_var_mean = np.exp(-self.x_var / 200) * 10
        self.y_var_std = np.abs(np.sin(self.y_var_mean))
        self.x_var_widths = np.ones_like(self.x_var) * 5

        self.y_var_mean_2 = np.exp(-self.x_var / 100) * 10
        self.y_var_std_2 = np.abs(np.sin(self.y_var_mean_2))

        self.test = VarVsVar(
            x_var=self.x_var,
            y_var_mean=self.y_var_mean,
            y_var_std=self.y_var_std,
            label=r"$10e^{-x/200}$",
            fill=False,
            is_marker=True,
        )
        self.test_2 = VarVsVar(
            x_var=self.x_var,
            y_var_mean=self.y_var_mean_2,
            y_var_std=self.y_var_std_2,
            x_var_widths=self.x_var_widths,
            label=r"$10e^{-x/100}$",
            fill=True,
            plot_y_std=False,
            is_marker=False,
        )

    def test_n_ratio_panels(self):
        """Check if ValueError is raised when we require more than 1 ratio panel."""
        with self.assertRaises(ValueError):
            VarVsVarPlot(
                n_ratio_panels=np.random.randint(2, 10),
            )

    def test_no_reference(self):
        """Check if ValueError is raised when plot ratios without reference."""
        test_plot = VarVsVarPlot(
            n_ratio_panels=1,
        )
        test_plot.add(self.test)
        test_plot.add(self.test_2)
        with self.assertRaises(ValueError):
            test_plot.plot_ratios()

    def test_overwrite_reference(self):
        """Check correct reference overwrite."""
        test_plot = VarVsVarPlot(
            n_ratio_panels=1,
        )
        test_plot.add(self.test, reference=True)
        test_plot.add(self.test_2, reference=True)

    def test_same_keys(self):
        """Check if KeyError is rased when we add VarVsVar object with existing key."""
        test_plot = VarVsVarPlot(
            n_ratio_panels=1,
        )
        test_plot.add(self.test, key="1")
        with self.assertRaises(KeyError):
            test_plot.add(self.test_2, key="1")

    def test_output_plot(self):
        """Test output plot."""
        # define the curves

        test_plot = VarVsVarPlot(
            ylabel=r"$\overline{N}_{trk}$",
            xlabel=r"$p_{T}$ [GeV]",
            grid=True,
            logy=False,
            atlas_second_tag="Unit test plot based on exponential decay.",
            n_ratio_panels=1,
            figsize=(9, 6),
        )
        test_plot.add(self.test, reference=True)
        test_plot.add(self.test_2, reference=False)

        test_plot.draw_hline(4)
        test_plot.draw()

        name = "test_var_vs_var.png"
        test_plot.savefig(f"{self.actual_plots_dir}/{name}")
        # Uncomment line below to update expected image
        # shutil.copy(f"{self.actual_plots_dir}/{name}", f"{self.expected_plots_dir}/{name}")
        self.assertEqual(
            None,
            compare_images(
                f"{self.actual_plots_dir}/{name}",
                f"{self.expected_plots_dir}/{name}",
                tol=5,
            ),
        )



"""Unit test script for the functions in hlplots/aux_results.py."""

from __future__ import annotations

import tempfile
import unittest
from pathlib import Path

import h5py
import numpy as np
from ftag.hdf5 import structured_from_dict

from puma.hlplots import AuxResults
from puma.hlplots.tagger import Tagger
from puma.utils import get_dummy_tagger_aux, logger, set_log_level

set_log_level(logger, "DEBUG")


class AuxResultsTestCase(unittest.TestCase):
    """Test class for the AuxResults class."""

    def test_add_duplicated(self):
        """Test empty string as model name."""
        dummy_tagger_1 = Tagger("dummy")
        dummy_tagger_2 = Tagger("dummy")
        results = AuxResults(sample="test")
        results.add(dummy_tagger_1)
        with self.assertRaises(KeyError):
            results.add(dummy_tagger_2)

    def test_add_2_taggers(self):
        """Test empty string as model name."""
        dummy_tagger_1 = Tagger("dummy")
        dummy_tagger_2 = Tagger("dummy_2")
        results = AuxResults(sample="test")
        results.add(dummy_tagger_1)
        results.add(dummy_tagger_2)
        self.assertEqual(
            list(results.taggers.keys()),
            ["dummy (dummy)", "dummy_2 (dummy_2)"],  # pylint: disable=W0212
        )

    def test_get_taggers(self):
        """Test empty string as model name."""
        dummy_tagger_1 = Tagger("dummy")
        dummy_tagger_2 = Tagger("dummy_2")
        results = AuxResults(sample="test")
        results.add(dummy_tagger_1)
        results.add(dummy_tagger_2)
        retrieved_dummy_tagger_2 = results["dummy_2 (dummy_2)"]
        self.assertEqual(retrieved_dummy_tagger_2.name, dummy_tagger_2.name)

    def test_load_taggers_from_file(self):
        """Test for AuxResults.load_taggers_from_file function."""
        np.random.default_rng(seed=16)
        fname = get_dummy_tagger_aux()[0]
        results = AuxResults(sample="test")
        taggers = [Tagger("GN2")]
        results.load_taggers_from_file(
            taggers,
            fname,
        )
        self.assertEqual(list(results.taggers.values()), taggers)

    def test_load_taggers_with_cuts(self):
        np.random.default_rng(seed=16)
        fname = get_dummy_tagger_aux()[0]
        cuts = [("eta", ">", 0)]
        tagger_cuts = [("pt", ">", 20)]
        results = AuxResults(sample="test")
        taggers = [Tagger("GN2", cuts=tagger_cuts)]
        results.load_taggers_from_file(
            taggers,
            fname,
            cuts=cuts,
        )
        self.assertEqual(list(results.taggers.values()), taggers)

    def test_load_taggers_with_cuts_override_perf_vars(self):
        rng = np.random.default_rng(seed=16)
        fname = get_dummy_tagger_aux()[0]
        cuts = [("eta", ">", 0)]
        tagger_cuts = [("pt", ">", 20)]
        results = AuxResults(sample="test")
        taggers = [Tagger("GN2", cuts=tagger_cuts)]
        results.load_taggers_from_file(
            taggers,
            fname,
            cuts=cuts,
            perf_vars={
                "pt": rng.exponential(100, size=9999),
                "eta": rng.normal(0, 1, size=9999),
            },
        )
        self.assertEqual(list(results.taggers.values()), taggers)

    def test_load_taggers_with_aux_perf_vars_eta(self):
        np.random.default_rng(seed=16)
        fname = get_dummy_tagger_aux()[0]
        results = AuxResults(
            sample="test",
            aux_perf_vars=["pt", "eta", "dphi"],
        )
        taggers = [Tagger("GN2")]
        results.load_taggers_from_file(taggers, fname)
        tagger_key = list(results.taggers.keys())[0]
        self.assertEqual(
            set(results.taggers[tagger_key].aux_perf_vars.keys()), {"pt", "eta", "dphi"}
        )

    def test_load_taggers_with_aux_perf_vars_deta(self):
        np.random.default_rng(seed=16)
        fname = get_dummy_tagger_aux()[0]
        results = AuxResults(
            sample="test",
            perf_vars=["eta"],
            aux_perf_vars=["pt", "deta", "dphi"],
        )
        taggers = [Tagger("GN2")]
        results.load_taggers_from_file(taggers, fname)
        tagger_key = list(results.taggers.keys())[0]
        self.assertEqual(
            set(results.taggers[tagger_key].aux_perf_vars.keys()), {"pt", "eta", "dphi"}
        )

    def test_add_taggers_keep_nan(self):
        # get mock file and add nans
        f = get_dummy_tagger_aux(size=500)[1]
        d = {}
        d["HadronConeExclTruthLabelID"] = f["jets"]["HadronConeExclTruthLabelID"]
        d["pt"] = f["jets"]["pt"]
        n_nans = np.random.choice(range(len(d["pt"])), 10, replace=False)
        d["pt"][n_nans] = np.nan
        jet_array = structured_from_dict(d)
        track_array = f["tracks"]
        with tempfile.TemporaryDirectory() as tmp_file:
            fname = Path(tmp_file) / "test.h5"
            with h5py.File(fname, "w") as f:
                f.create_dataset("jets", data=jet_array)
                f.create_dataset("tracks", data=track_array)
            results = AuxResults(sample="test", remove_nan=False)
            with self.assertRaises(ValueError):
                results.load_taggers_from_file([Tagger("GN2")], fname)

    def test_add_taggers_remove_nan(self):
        # get mock file and add nans
        f = get_dummy_tagger_aux(size=500)[1]
        d = {}
        d["HadronConeExclTruthLabelID"] = f["jets"]["HadronConeExclTruthLabelID"]
        d["pt"] = f["jets"]["pt"]
        n_nans = np.random.choice(range(len(d["pt"])), 10, replace=False)
        d["pt"][n_nans] = np.nan
        jet_array = structured_from_dict(d)
        track_array = f["tracks"]
        with tempfile.TemporaryDirectory() as tmp_file:
            fname = Path(tmp_file) / "test.h5"
            with h5py.File(fname, "w") as f:
                f.create_dataset("jets", data=jet_array)
                f.create_dataset("tracks", data=track_array)
            results = AuxResults(sample="test", remove_nan=True)
            with self.assertLogs("puma", "WARNING") as cm:
                results.load_taggers_from_file([Tagger("GN2")], fname)
            self.assertEqual(
                cm.output,
                [f"WARNING:puma:{len(n_nans)} NaN values found in loaded data." " Removing them."],
            )


class AuxResultsPlotsTestCase(unittest.TestCase):
    """Test class for the AuxResults class running plots."""

    def setUp(self) -> None:
        """Set up for unit tests."""
        f = get_dummy_tagger_aux()[1]
        dummy_tagger = Tagger("GN2")
        dummy_tagger.labels = np.array(
            f["jets"]["HadronConeExclTruthLabelID"],
            dtype=[("HadronConeExclTruthLabelID", "i4")],
        )
        dummy_tagger.aux_scores = {
            "vertexing": f["tracks"]["GN2_aux_VertexIndex"],
            "track_origin": f["tracks"]["GN2_aux_TrackOrigin"],
        }
        dummy_tagger.aux_labels = {
            "vertexing": f["tracks"]["ftagTruthVertexIndex"],
            "track_origin": f["tracks"]["ftagTruthOriginLabel"],
        }
        dummy_tagger.perf_vars = {"pt": f["jets"]["pt"]}
        dummy_tagger.aux_perf_vars = {
            "pt": f["tracks"]["pt"],
            "eta": f["tracks"]["eta"],
            "dphi": f["tracks"]["dphi"],
        }
        dummy_tagger.label = "dummy tagger"
        dummy_tagger_no_aux = Tagger("GN2_NoAux", aux_tasks=[])
        dummy_tagger_no_aux.perf_vars = {"pt": f["jets"]["pt"]}
        dummy_tagger_no_aux.label = "dummy tagger no aux"
        self.dummy_tagger = dummy_tagger
        self.dummy_tagger_no_aux = dummy_tagger_no_aux

    def assertIsFile(self, path: str):
        """Check for file to exist.
        Taken from https://stackoverflow.com/a/59198749/10896585.

        Parameters
        ----------
        path : str
            Path to file.

        Raises
        ------
        AssertionError
            if file does not exist
        """
        if not Path(path).resolve().is_file():
            raise AssertionError(f"File does not exist: {path}")

    def test_plot_trackorigin_cm_minmal(self):
        self.dummy_tagger.reference = True
        with tempfile.TemporaryDirectory() as tmp_file:
            auxresults = AuxResults(sample="test", output_dir=tmp_file)
            auxresults.add(self.dummy_tagger)
            auxresults.plot_track_origin_confmat()
            self.assertIsFile(auxresults.get_filename(self.dummy_tagger.name + "_trackOrigin_cm"))

    def test_plot_trackorigin_cm_full(self):
        self.dummy_tagger.reference = True
        with tempfile.TemporaryDirectory() as tmp_file:
            auxresults = AuxResults(sample="test", output_dir=tmp_file)
            auxresults.add(self.dummy_tagger)
            auxresults.plot_track_origin_confmat(minimal_plot=False)
            self.assertIsFile(auxresults.get_filename(self.dummy_tagger.name + "_trackOrigin_cm"))

    def test_plot_var_vtx_perf_bjets(self):
        """Test that png files are being created for tagger with aux tasks."""
        self.dummy_tagger.reference = True
        with tempfile.TemporaryDirectory() as tmp_file:
            auxresults = AuxResults(sample="test", output_dir=tmp_file)
            auxresults.add(self.dummy_tagger)
            auxresults.plot_var_vtx_perf(vtx_flavours=["bjets"])
            self.assertIsFile(auxresults.get_filename("bjets_vtx_eff_vs_pt"))
            self.assertIsFile(auxresults.get_filename("bjets_vtx_purity_vs_pt"))
            self.assertIsFile(auxresults.get_filename("bjets_vtx_trk_eff_vs_pt"))
            self.assertIsFile(auxresults.get_filename("bjets_vtx_trk_purity_vs_pt"))

    def test_plot_var_vtx_perf_ujets(self):
        """Test that png files are being created for tagger with aux tasks."""
        self.dummy_tagger.reference = True
        with tempfile.TemporaryDirectory() as tmp_file:
            auxresults = AuxResults(sample="test", output_dir=tmp_file)
            auxresults.add(self.dummy_tagger)
            auxresults.plot_var_vtx_perf(no_vtx_flavours=["ujets"])
            self.assertIsFile(auxresults.get_filename("ujets_vtx_fakes_vs_pt"))

    def test_plot_var_vtx_perf_alljets(self):
        """Test that png files are being created for tagger with aux tasks."""
        self.dummy_tagger.reference = True
        with tempfile.TemporaryDirectory() as tmp_file:
            auxresults = AuxResults(sample="test", output_dir=tmp_file)
            auxresults.add(self.dummy_tagger)
            auxresults.plot_var_vtx_perf(vtx_flavours=["bjets", "cjets"], no_vtx_flavours=["ujets"])
            self.assertIsFile(auxresults.get_filename("bjets_vtx_eff_vs_pt"))
            self.assertIsFile(auxresults.get_filename("bjets_vtx_purity_vs_pt"))
            self.assertIsFile(auxresults.get_filename("bjets_vtx_trk_eff_vs_pt"))
            self.assertIsFile(auxresults.get_filename("bjets_vtx_trk_purity_vs_pt"))
            self.assertIsFile(auxresults.get_filename("cjets_vtx_eff_vs_pt"))
            self.assertIsFile(auxresults.get_filename("cjets_vtx_purity_vs_pt"))
            self.assertIsFile(auxresults.get_filename("cjets_vtx_trk_eff_vs_pt"))
            self.assertIsFile(auxresults.get_filename("cjets_vtx_trk_purity_vs_pt"))
            self.assertIsFile(auxresults.get_filename("ujets_vtx_fakes_vs_pt"))

    def test_plot_var_vtx_noaux(self):
        """Test that error is raised if no tagger with vertexing is added."""
        self.dummy_tagger.reference = True
        with tempfile.TemporaryDirectory() as tmp_file:
            auxresults = AuxResults(sample="test", output_dir=tmp_file)
            auxresults.add(self.dummy_tagger_no_aux)
            with self.assertRaises(ValueError):
                auxresults.plot_var_vtx_perf(vtx_flavours=["bjets"])

    def test_plot_var_vtx_perf_no_flavours(self):
        """Test vertexing performance when no jet flavours are specified."""
        self.dummy_tagger.reference = True
        with tempfile.TemporaryDirectory() as tmp_file:
            auxresults = AuxResults(sample="test", output_dir=tmp_file)
            auxresults.add(self.dummy_tagger)
            with self.assertRaises(ValueError):
                auxresults.plot_var_vtx_perf()

    def test_plot_var_vtx_perf_empty(self):
        """Test vertexing performance function with empty data."""
        self.dummy_tagger_no_aux.reference = True
        with tempfile.TemporaryDirectory() as tmp_file:
            auxresults = AuxResults(sample="test", output_dir=tmp_file)
            auxresults.add(self.dummy_tagger_no_aux)
            with self.assertRaises(ValueError):
                auxresults.plot_var_vtx_perf(vtx_flavours=["bjets"])

    def test_plot_var_vtx_perf_bjets_inclusive_vertexing(self):
        """Test that png files are being created for tagger with inclusive vertexing enabled."""
        self.dummy_tagger.reference = True
        with tempfile.TemporaryDirectory() as tmp_file:
            auxresults = AuxResults(sample="test", output_dir=tmp_file)
            auxresults.add(self.dummy_tagger)
            auxresults.plot_var_vtx_perf(vtx_flavours=["bjets"], incl_vertexing=True)
            self.assertIsFile(auxresults.get_filename("bjets_vtx_eff_vs_pt", suffix="incl"))
            self.assertIsFile(auxresults.get_filename("bjets_vtx_purity_vs_pt", suffix="incl"))
            self.assertIsFile(auxresults.get_filename("bjets_vtx_trk_eff_vs_pt", suffix="incl"))
            self.assertIsFile(auxresults.get_filename("bjets_vtx_trk_purity_vs_pt", suffix="incl"))

    def test_plot_var_vtx_perf_ujets_inclusive_vertexing(self):
        """Test that png files are being created for tagger with inclusive vertexing enabled."""
        self.dummy_tagger.reference = True
        with tempfile.TemporaryDirectory() as tmp_file:
            auxresults = AuxResults(sample="test", output_dir=tmp_file)
            auxresults.add(self.dummy_tagger)
            auxresults.plot_var_vtx_perf(no_vtx_flavours=["ujets"], incl_vertexing=True)
            self.assertIsFile(auxresults.get_filename("ujets_vtx_fakes_vs_pt", suffix="incl"))

    def test_plot_var_mass_inclusive_vertexing(self):
        """Test that png files are being created for mass reconstruction."""
        self.dummy_tagger.reference = True
        with tempfile.TemporaryDirectory() as tmp_file:
            auxresults = AuxResults(
                sample="test",
                aux_perf_vars=["pt", "eta", "dphi"],
                output_dir=tmp_file,
            )
            auxresults.add(self.dummy_tagger)
            auxresults.plot_vertex_mass(vtx_flavours=["bjets"], incl_vertexing=True)
            self.assertIsFile(auxresults.get_filename("bjets_sv_mass", suffix="incl"))
            self.assertIsFile(auxresults.get_filename("bjets_sv_mass_diff"))

    def test_plot_var_mass_exclusive_vertexing(self):
        """Test that png files are being created for mass reconstruction."""
        self.dummy_tagger.reference = True
        with tempfile.TemporaryDirectory() as tmp_file:
            auxresults = AuxResults(
                sample="test",
                aux_perf_vars=["pt", "eta", "dphi"],
                output_dir=tmp_file,
            )
            auxresults.add(self.dummy_tagger)
            auxresults.plot_vertex_mass(vtx_flavours=["bjets"], incl_vertexing=False)
            self.assertIsFile(auxresults.get_filename("bjets_sv_mass", suffix="excl"))



"""Efficiency plots vs. specific variable."""

from __future__ import annotations

from typing import ClassVar

import numpy as np

from puma.metrics import eff_err, rej_err
from puma.utils import logger
from puma.utils.histogram import save_divide
from puma.var_vs_var import VarVsVar, VarVsVarPlot


class VarVsEff(VarVsVar):  # pylint: disable=too-many-instance-attributes
    """Class for efficiency vs. variable plot."""

    def __init__(
        self,
        x_var_sig: np.ndarray,
        disc_sig: np.ndarray,
        x_var_bkg: np.ndarray = None,
        disc_bkg: np.ndarray = None,
        bins=10,
        working_point: float | None = None,
        disc_cut=None,
        flat_per_bin: bool = False,
        key: str | None = None,
        **kwargs,
    ) -> None:
        """Initialise properties of roc curve object.

        Parameters
        ----------
        x_var_sig : np.ndarray
            Values for x-axis variable for signal
        disc_sig : np.ndarray
            Discriminant values for signal
        x_var_bkg : np.ndarray, optional
            Values for x-axis variable for background, by default None
        disc_bkg : np.ndarray, optional
            Discriminant values for background, by default None
        bins : int or sequence of scalars, optional
            If bins is an int, it defines the number of equal-width bins in the
            given range (10, by default). If bins is a sequence, it defines a
            monotonically increasing array of bin edges, including the
            rightmost edge, allowing for non-uniform bin widths, by default 10
        working_point : float, optional
            Working point, by default None
        disc_cut : float or  sequence of floats, optional
            Cut value for discriminant, if it is a sequence it has to have the same
            length as number of bins, by default None
        flat_per_bin : bool, optional
            If True and no `disc_cut` is given the signal efficiency is held constant
            in each bin, by default False
        key : str, optional
            Identifier for the curve e.g. tagger, by default None
        **kwargs : kwargs
            Keyword arguments passed to `PlotLineObject`

        Raises
        ------
        ValueError
            If provided options are not compatible with each other
        """
        if len(x_var_sig) != len(disc_sig):
            raise ValueError(
                f"Length of `x_var_sig` ({len(x_var_sig)}) and `disc_sig` "
                f"({len(disc_sig)}) have to be identical."
            )
        if x_var_bkg is not None and len(x_var_bkg) != len(disc_bkg):
            raise ValueError(
                f"Length of `x_var_bkg` ({len(x_var_bkg)}) and `disc_bkg` "
                f"({len(disc_bkg)}) have to be identical."
            )
        # checking that the given options are compatible
        # could also think about porting it to a class function insted of passing
        # the arguments to init e.g. `set_method`
        if working_point is None and disc_cut is None:
            raise ValueError("Either `wp` or `disc_cut` needs to be specified.")
        if flat_per_bin:
            if disc_cut is not None:
                raise ValueError(
                    "You cannot specify `disc_cut` when `flat_per_bin` is set to True."
                )
            if working_point is None:
                raise ValueError(
                    "You need to specify a working point `wp`, when `flat_per_bin` is"
                    " set to True."
                )
        self.x_var_sig = np.array(x_var_sig)
        self.disc_sig = np.array(disc_sig)
        self.x_var_bkg = None if x_var_bkg is None else np.array(x_var_bkg)
        self.disc_bkg = None if disc_bkg is None else np.array(disc_bkg)
        self.working_point = working_point
        self.disc_cut = disc_cut
        self.flat_per_bin = flat_per_bin
        # Binning related variables
        self.n_bins = None
        self.bn_edges = None
        self.x_bin_centres = None
        self.bin_widths = None
        self.n_bins = None
        # Binned distributions
        self.bin_indices_sig = None
        self.disc_binned_sig = None
        self.bin_indices_bkg = None
        self.disc_binned_bkg = None

        self._set_bin_edges(bins)

        if disc_cut is not None:
            if working_point is not None:
                raise ValueError("You cannot specify `disc_cut` when providing `wp`.")
            if isinstance(disc_cut, (list, np.ndarray)) and self.n_bins != len(disc_cut):
                raise ValueError(
                    "`disc_cut` has to be a float or has to have the same length as"
                    " number of bins."
                )
        self._apply_binning()
        self._get_disc_cuts()

        VarVsVar.__init__(
            self,
            x_var=self.x_bin_centres,
            y_var_mean=np.zeros_like(self.x_bin_centres),
            y_var_std=np.zeros_like(self.x_bin_centres),
            x_var_widths=2 * self.bin_widths,
            key=key,
            fill=True,
            plot_y_std=False,
            **kwargs,
        )
        self.inverse_cut = False

    def _set_bin_edges(self, bins):
        """Calculate bin edges, centres and width and save them as class variables.

        Parameters
        ----------
        bins : int or sequence of scalars
            If bins is an int, it defines the number of equal-width bins in the given
            range. If bins is a sequence, it defines a monotonically increasing array of
            bin edges, including the rightmost edge, allowing for non-uniform bin
            widths.
        """
        logger.debug("Calculating binning.")
        if isinstance(bins, int):
            # With this implementation, the data point with x=xmax will be added to the
            # overflow bin.
            xmin, xmax = np.amin(self.x_var_sig), np.amax(self.x_var_sig)
            if self.x_var_bkg is not None:
                xmin = min(xmin, np.amin(self.x_var_bkg))
                xmax = max(xmax, np.amax(self.x_var_bkg))
            # increasing xmax slightly to inlcude largest value due to hehavior of
            # np.digitize
            xmax *= 1 + 1e-5
            self.bin_edges = np.linspace(xmin, xmax, bins + 1)
        elif isinstance(bins, (list, np.ndarray)):
            self.bin_edges = np.array(bins)
        logger.debug("Retrieved bin edges %s}", self.bin_edges)
        # Get the bins for the histogram
        self.x_bin_centres = (self.bin_edges[:-1] + self.bin_edges[1:]) / 2.0
        self.bin_widths = (self.bin_edges[1:] - self.bin_edges[:-1]) / 2.0
        self.n_bins = self.bin_edges.size - 1
        logger.debug("N bins: %i", self.n_bins)

    def _apply_binning(self):
        """Get binned distributions for the signal and background."""
        logger.debug("Applying binning.")
        self.bin_indices_sig = np.digitize(self.x_var_sig, self.bin_edges)
        if np.all(self.bin_indices_sig == 0):
            logger.error("All your signal is in the underflow bin. Check your input.")
            # retrieve for each bin the part of self.disc_sig corresponding to this bin
            # and put them in a list
        self.disc_binned_sig = [
            self.disc_sig[np.where(self.bin_indices_sig == x)[0]]
            for x in range(1, len(self.bin_edges))
        ]
        if self.x_var_bkg is not None:
            self.bin_indices_bkg = np.digitize(self.x_var_bkg, self.bin_edges)
            self.disc_binned_bkg = [
                self.disc_bkg[np.where(self.bin_indices_bkg == x)[0]]
                for x in range(1, len(self.bin_edges))
            ]

    def _get_disc_cuts(self):
        """Retrieve cut values on discriminant. If `disc_cut` is not given, retrieve
        cut values from the working point.
        """
        logger.debug("Calculate discriminant cut.")
        if isinstance(self.disc_cut, (float, int)):
            self.disc_cut = [self.disc_cut] * self.n_bins
        elif isinstance(self.disc_cut, (list, np.ndarray)):
            self.disc_cut = self.disc_cut
        elif self.flat_per_bin:
            self.disc_cut = [
                np.percentile(x, (1 - self.working_point) * 100) for x in self.disc_binned_sig
            ]
        else:
            self.disc_cut = [
                np.percentile(self.disc_sig, (1 - self.working_point) * 100)
            ] * self.n_bins
        logger.debug("Discriminant cut: %.3f", self.disc_cut)

    def efficiency(self, arr: np.ndarray, cut: float):
        """Calculate efficiency and the associated error.

        Parameters
        ----------
        arr : np.ndarray
            Array with discriminants
        cut : float
            Cut value

        Returns
        -------
        float
            Efficiency
        float
            Efficiency error
        """
        if len(arr) == 0:
            return 0, 0
        eff = sum(arr < cut) / len(arr) if self.inverse_cut else sum(arr > cut) / len(arr)
        eff_error = eff_err(eff, len(arr))
        return eff, eff_error

    def rejection(self, arr: np.ndarray, cut: float):
        """Calculate rejection and the associated error.

        Parameters
        ----------
        arr : np.ndarray
            Array with discriminants
        cut : float
            Cut value

        Returns
        -------
        float
            Rejection
        float
            Rejection error
        """
        if self.inverse_cut:
            rej = save_divide(len(arr), sum(arr < cut), default=np.inf)
        else:
            rej = save_divide(len(arr), sum(arr > cut), default=np.inf)
        if rej == np.inf:
            logger.warning("Your rejection is infinity -> setting it to np.nan.")
            return np.nan, np.nan
        rej_error = rej_err(rej, len(arr))
        return rej, rej_error

    @property
    def bkg_eff_sig_err(self):
        """Calculate signal efficiency per bin, assuming a flat background per
        bin. This results in returning the signal efficiency per bin, but the
        background error per bin.
        """
        logger.debug("Calculating signal efficiency.")
        eff = np.array(list(map(self.efficiency, self.disc_binned_bkg, self.disc_cut)))[:, 0]
        err = np.array(list(map(self.efficiency, self.disc_binned_sig, self.disc_cut)))[:, 1]
        logger.debug("Retrieved signal efficiencies: %s", eff)
        return eff, err

    @property
    def sig_eff(self):
        """Calculate signal efficiency per bin.

        Returns
        -------
        np.ndarray
            Efficiency
        np.ndarray
            Efficiency_error
        """
        logger.debug("Calculating signal efficiency.")
        eff = list(map(self.efficiency, self.disc_binned_sig, self.disc_cut))
        logger.debug("Retrieved signal efficiencies: %s", eff)
        return np.array(eff)[:, 0], np.array(eff)[:, 1]

    @property
    def bkg_eff(self):
        """Calculate background efficiency per bin.

        Returns
        -------
        np.ndarray
            Efficiency
        np.ndarray
            Efficiency_error
        """
        logger.debug("Calculating background efficiency.")
        eff = list(map(self.efficiency, self.disc_binned_bkg, self.disc_cut))
        logger.debug("Retrieved background efficiencies: %.2f", eff)
        return np.array(eff)[:, 0], np.array(eff)[:, 1]

    @property
    def sig_rej(self):
        """Calculate signal rejection per bin.

        Returns
        -------
        np.ndarray
            Rejection
        np.ndarray
            Rejection_error
        """
        logger.debug("Calculating signal rejection.")
        rej = list(map(self.rejection, self.disc_binned_sig, self.disc_cut))
        logger.debug("Retrieved signal rejections: %.1f", rej)
        return np.array(rej)[:, 0], np.array(rej)[:, 1]

    @property
    def bkg_rej(self):
        """Calculate background rejection per bin.

        Returns
        -------
        np.ndarray
            Rejection
        np.ndarray
            Rejection_error
        """
        logger.debug("Calculating background rejection.")
        rej = list(map(self.rejection, self.disc_binned_bkg, self.disc_cut))
        logger.debug("Retrieved background rejections: %s", rej)
        return np.array(rej)[:, 0], np.array(rej)[:, 1]

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return (
                np.all(self.x_var_sig == other.x_var_sig)
                and np.all(self.disc_sig == other.disc_sig)
                and np.all(self.x_var_bkg == other.x_var_bkg)
                and np.all(self.disc_bkg == other.disc_bkg)
                and np.all(self.bn_edges == other.bn_edges)
                and self.working_point == other.working_point
                and np.all(self.disc_cut == other.disc_cut)
                and self.flat_per_bin == other.flat_per_bin
                and self.key == other.key
            )
        return False

    def get(self, mode: str, inverse_cut: bool = False):
        """Wrapper around rejection and efficiency functions.

        Parameters
        ----------
        mode : str
            Can be "sig_eff", "bkg_eff", "sig_rej", "bkg_rej", or
            "bkg_eff_sig_err"
        inverse_cut : bool, optional
            Inverts the discriminant cut, which will yield the efficiency or rejection
            of the jets not passing the working point, by default False

        Returns
        -------
        np.ndarray
            Rejection or efficiency depending on `mode` value
        np.ndarray
            Rejection or efficiency error depending on `mode` value

        Raises
        ------
        ValueError
            If mode not supported
        """
        self.inverse_cut = inverse_cut
        if mode == "sig_eff":
            return self.sig_eff
        if mode == "bkg_eff":
            return self.bkg_eff
        if mode == "sig_rej":
            return self.sig_rej
        if mode == "bkg_rej":
            return self.bkg_rej
        if mode == "bkg_eff_sig_err":
            return self.bkg_eff_sig_err
        # setting class variable again to False
        self.inverse_cut = False
        raise ValueError(
            f"The selected mode {mode} is not supported. Use one of the following:"
            f" {VarVsEffPlot.mode_options}."
        )


class VarVsEffPlot(VarVsVarPlot):  # pylint: disable=too-many-instance-attributes
    """var_vs_eff plot class."""

    mode_options: ClassVar[list[str]] = [
        "sig_eff",
        "bkg_eff",
        "sig_rej",
        "bkg_rej",
        "bkg_eff_sig_err",
    ]

    def __init__(self, mode, grid: bool = False, **kwargs) -> None:
        """var_vs_eff plot properties.

        Parameters
        ----------
        mode : str
            Defines which quantity is plotted, the following options ar available:
                sig_eff - Plots signal efficiency vs. variable, with statistical error
                    on N signal per bin
                bkg_eff - Plots background efficiency vs. variable, with statistical
                    error on N background per bin
                sig_rej - Plots signal rejection vs. variable, with statistical error
                    on N signal per bin
                bkg_rej - Plots background rejection vs. variable, with statistical
                    error on N background per bin
                bkg_eff_sig_err - Plots background efficiency vs. variable, with
                    statistical error on N signal per bin.
        grid : bool, optional
            Set the grid for the plots.
        **kwargs : kwargs
            Keyword arguments from `puma.PlotObject`

        Raises
        ------
        ValueError
            If incompatible mode given or more than 1 ratio panel requested
        """
        super().__init__(grid=grid, **kwargs)
        if mode not in self.mode_options:
            raise ValueError(
                f"The selected mode {mode} is not supported. Use one of the following: "
                f"{self.mode_options}."
            )
        self.mode = mode

    def _setup_curves(self):
        for key in self.add_order:
            elem = self.plot_objects[key]
            y_value, y_error = elem.get(self.mode, inverse_cut=self.inverse_cut)
            elem.y_var_mean = y_value
            elem.y_var_std = y_error

    def apply_modified_atlas_second_tag(
        self,
        signal,
        working_point=None,
        disc_cut=None,
        flat_per_bin=False,
    ):
        """Modifies the atlas_second_tag to include info on the type of p-eff plot
        being displayed.
        """
        if working_point:
            mid_str = f"{round(working_point * 100, 3)}% " + signal.eff_str
        elif disc_cut:
            mid_str = rf"$D_{{{signal.name.rstrip('jets')}}}$ > {disc_cut}"
        tag = f"Flat {mid_str} per bin" if flat_per_bin else f"{mid_str}"
        if self.atlas_second_tag:
            self.atlas_second_tag = f"{self.atlas_second_tag}\n{tag}"
        else:
            self.atlas_second_tag = tag

    def plot(self, **kwargs):
        """Plotting curves.

        Parameters
        ----------
        **kwargs: kwargs
            Keyword arguments passed to plt.axis.errorbar

        Returns
        -------
        Line2D
            matplotlib Line2D object
        """
        logger.debug("Plotting curves with mode %s", self.mode)
        self._setup_curves()
        return super().plot(**kwargs)
